# connect_wifi.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:44 PM

"""Connect to the Wifi AP of a GoPro camera."""

import argparse
import asyncio
import logging
from typing import Optional

from rich.console import Console

from open_gopro import WirelessGoPro
from open_gopro.util import add_cli_args_and_parse, ainput
from open_gopro.util.logger import set_stream_logging_level, setup_logging

console = Console()


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Connect to a GoPro camera's Wifi Access Point.")
    return add_cli_args_and_parse(parser)


async def main(args: argparse.Namespace) -> None:
    setup_logging(__name__, args.log)
    gopro: Optional[WirelessGoPro] = None

    async with WirelessGoPro(
        args.identifier, host_wifi_interface=args.wifi_interface, host_sudo_password=args.password
    ) as gopro:
        # Now we only want errors
        set_stream_logging_level(logging.ERROR)

        console.print("\n\n🎆🎇✨ Success!! Wifi AP is connected 📡\n")
        console.print("Send commands as per https://gopro.github.io/OpenGoPro/http")

        await ainput("[blue]Press enter to disconnect Wifi and exit...", console.print)
        console.print("Exiting...")

    if gopro:
        await gopro.close()


# Needed for poetry scripts defined in pyproject.toml
def entrypoint() -> None:
    asyncio.run(main(parse_arguments()))


if __name__ == "__main__":
    entrypoint()



# cohn.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Oct 24 19:08:07 UTC 2023

"""Entrypoint for configuring and demonstrating Camera On the Home Network (COHN)."""

from __future__ import annotations

import argparse
import asyncio
import textwrap
from pathlib import Path

from rich.console import Console

from open_gopro import WirelessGoPro
from open_gopro.util import add_cli_args_and_parse
from open_gopro.util.logger import setup_logging

console = Console()  # rich consoler printer

COHN_CURL_CMD_TEMPLATE = (
    r"""curl --insecure -v -u 'gopro:{password}' --cacert cohn.crt 'https://{ip_addr}/gopro/camera/state'"""
)


def dump_cohn_collateral(gopro: WirelessGoPro) -> None:
    """Print the COHN credentials and write the certificate to a file.

    Args:
        gopro (WirelessGoPro): gopro to retrieve the credentials from
    """
    assert gopro.cohn.credentials
    console.print(
        f"Sample curl command: {COHN_CURL_CMD_TEMPLATE.format(
        password=gopro.cohn.credentials.password,
        ip_addr=gopro.cohn.credentials.ip_address,)}"
    )
    with open("cohn.crt", "w") as f:
        f.write(gopro.cohn.credentials.certificate)


async def main(args: argparse.Namespace) -> None:
    logger = setup_logging(__name__, args.log)

    gopro: WirelessGoPro | None = None
    try:
        # If we weren't explicitly passed an identifier, we will try to find the first camera available
        if identifier := args.identifier:
            console.print(f"Attempting to directly communicate via COHN to GoPro {identifier}")
        else:
            # Start with Wifi Disabled (i.e. don't allow camera in AP mode) and only connect BLE to check COHN status / provision
            async with WirelessGoPro(
                target=args.identifier,
                interfaces={WirelessGoPro.Interface.BLE},
                cohn_db=args.db,
            ) as gopro:
                if await gopro.cohn.is_configured:
                    console.print("COHN is already configured.")
                else:
                    if not args.ssid or not args.password:
                        raise ValueError("COHN needs to be provisioned but you didn't pass SSID or password.")
                    await gopro.access_point.connect(args.ssid, args.password)
                    await gopro.cohn.configure(force_reprovision=True)

                console.print("[blue]COHN is ready for communication. Dropping the BLE connection.")
                dump_cohn_collateral(gopro)

            identifier = gopro.identifier
        # Now create an object with only COHN
        async with WirelessGoPro(
            target=identifier,
            interfaces={WirelessGoPro.Interface.COHN},
            cohn_db=args.db,
        ) as gopro:
            # Prove we can communicate via the COHN HTTP channel without a BLE or Wifi connection
            assert (await gopro.http_command.get_camera_state()).ok
            console.print("Successfully communicated via COHN!!")
            dump_cohn_collateral(gopro)

    except Exception as e:  # pylint: disable = broad-except
        logger.error(repr(e))

    if gopro:
        await gopro.close()


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """\
        Open GoPro Camera on the Home Network (COHN) Utility
        ----------------------------------------------------
        This utility is used to configure and demonstrate the Camera On the Home
        Network (COHN) feature of GoPro cameras. COHN allows the camera to be accessed
        over a local network using HTTP, without the need for a direct Wi-Fi or Bluetooth
        connection.

        There are three main modes of operation that are chosen based on the arguments
        passed to this script:

        1. COHN Provisioning: If the camera is not already provisioned for COHN, you can
           provide the SSID (--ssid) and password (--password) of the Wi-Fi network to
           which the camera should connect. The script will then (re)configure the camera
           for COHN using these credentials.

        2. COHN Communication without initial BLE check: If the camera is already
           provisioned for COHN, you can directly communicate with it using the COHN
           HTTP channel. This is done by passing the camera's identifier (--identifier)
           which is the trailing 4 digits of the serial number.

        3. COHN Communication with initial BLE check: If you don't provide an identifier,
           the script will first attempt to find the first available camera using BLE. If
           it finds a camera that is already provisioned for COHN, it will connect to it
           and demonstrate COHN communication. If it finds a camera that is not provisioned
           for COHN, it will return an error message and exit if you didn't pass SSID and
           password. If you pass SSID and password, this is the same as the first mode of
           operation.

        Therefore, the general procedure for a freshly factory-reset camera is:
            1. Use operation mode 1 to provision COHN
            2. Use operation mode 2 to communicate with (only) COHN
        """
        ),
    )
    parser.add_argument(
        "--ssid",
        type=str,
        help="WiFi SSID to connect. Mutually inclusive with --password.",
        default=None,
    )
    parser.add_argument(
        "--password",
        type=str,
        help="Password of WiFi SSID. Mutually inclusive with --ssid.",
        default=None,
    )
    parser.add_argument(
        "--db",
        type=Path,
        help="path to COHN database file. Defaults to cohn_db.json",
        default=Path("cohn_db.json"),
    )
    args = add_cli_args_and_parse(parser, wifi=False)
    if args.ssid and not args.password:
        parser.error("The --ssid argument requires --password to also be set.")
    return args


# Needed for poetry scripts defined in pyproject.toml
def entrypoint() -> None:
    asyncio.run(main(parse_arguments()))


if __name__ == "__main__":
    entrypoint()



# controller.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""BLE Controller Interface Definition."""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Callable, Generic, Optional, Pattern, TypeVar

from .services import BleUUID, GattDB, UUIDs

logger = logging.getLogger(__name__)

BleDevice = TypeVar("BleDevice")
BleHandle = TypeVar("BleHandle")
NotiHandlerType = Callable[[int, bytearray], None]
DisconnectHandlerType = Callable[[BleDevice], None]


class BLEController(ABC, Generic[BleDevice, BleHandle]):
    """Interface definition for a BLE driver to be used by GoPro."""

    def __init__(self, exception_handler: Optional[Callable] = None) -> None:
        self._exception_handler = exception_handler

    @abstractmethod
    async def read(self, handle: BleHandle, uuid: BleUUID) -> bytes:
        """Read a bytestream response from a BleUUID.

        Args:
            handle (BleHandle): handle to pair to
            uuid (BleUUID): BleUUID to read from

        Returns:
            bytes: response
        """

    @abstractmethod
    async def write(self, handle: BleHandle, uuid: BleUUID, data: bytes) -> None:
        """Write a bytestream to a BleUUID.

        Args:
            handle (BleHandle): handle to pair to
            uuid (BleUUID): BleUUID to write to
            data (bytes): bytestream to write
        """

    @abstractmethod
    async def scan(self, token: Pattern, timeout: int = 5, service_uuids: list[BleUUID] | None = None) -> BleDevice:
        """Scan BLE device with a regex in it's device name.

        Args:
            token (Pattern): Regex to scan for
            timeout (int): Time to scan (in seconds) before considering scanning as failed. Defaults to 5.
            service_uuids (list[BleUUID] | None): The list of BleUUID's to filter on. Defaults to None.

        Returns:
            BleDevice: discovered device (shall not be multiple devices)
        """

    @abstractmethod
    async def connect(self, disconnect_cb: DisconnectHandlerType, device: BleDevice, timeout: int = 15) -> BleHandle:
        """Connect to a BLE device.

        Args:
            disconnect_cb (DisconnectHandlerType): function to call when disconnect is received
            device (BleDevice): device to connect to
            timeout (int): How long to attempt connecting before giving up. Defaults to 15.

        Returns:
            BleHandle: handle that has been connected to
        """

    @abstractmethod
    async def pair(self, handle: BleHandle) -> None:
        """Pair to an already connected handle.

        Args:
            handle (BleHandle): handle to pair to
        """

    @abstractmethod
    async def enable_notifications(self, handle: BleHandle, handler: NotiHandlerType) -> None:
        """Enable notifications for all notifiable characteristics.

        The handler is used to register for notifications. It will be called when a a notification
        is received.

        Args:
            handle (BleHandle): handle to enable notifications on
            handler (NotiHandlerType): notification handler
        """

    @abstractmethod
    async def discover_chars(self, handle: BleHandle, uuids: type[UUIDs] | None = None) -> GattDB:
        """Discover all characteristics for a connected handle.

        By default, the BLE controller only knows Spec-Defined BleUUID's so any additional BleUUID's should
        be passed in with the uuids argument

        Args:
            handle (BleHandle): BLE handle to discover for
            uuids (type[UUIDs] | None): Additional BleUUID information to use when building the Gatt Database.
                Defaults to None.

        Returns:
            GattDB: Gatt Database
        """

    @abstractmethod
    async def disconnect(self, handle: BleHandle) -> None:
        """Terminate the BLE connection.

        Args:
            handle (BleHandle): handle to disconnect from
        """



# logger.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu Aug 24 17:08:14 UTC 2023

"""Logger abstraction above default python logging"""

from __future__ import annotations

import http.client as http_client
import logging
from pathlib import Path
from typing import Any, Final

from rich import traceback
from rich.logging import RichHandler


class Logger:
    """A singleton class to manage logging for the Open GoPro internal modules

    Attributes:
        ARROW_HEAD_COUNT (Final[int]): Length of leading arrow
        ARROW_TAIL_COUNT (Final[int]): Length of trailing arrow

    Args:
        logger (logging.Logger): input logger that will be modified and then returned
        output (Path | None): Path of log file for file stream handler. If not set, will not log to file.
        modules (list[str] | None): Optional override of modules / levels. Will be merged into default modules.
    """

    _instances: dict[type[Logger], Logger] = {}
    ARROW_HEAD_COUNT: Final[int] = 8
    ARROW_TAIL_COUNT: Final[int] = 14

    def __new__(cls, *_: Any) -> Any:  # noqa https://github.com/PyCQA/pydocstyle/issues/515
        if cls not in cls._instances:
            c = object.__new__(cls)
            cls._instances[cls] = c
            return c
        raise RuntimeError("The logger can only be setup once and this should be done at the top level.")

    def __init__(
        self,
        logger: logging.Logger,
        output: Path | None = None,
        modules: list[str] | None = None,
    ) -> None:
        self.modules = [
            "open_gopro.gopro_base",
            "open_gopro.gopro_wired",
            "open_gopro.gopro_wireless",
            "open_gopro.api.builders",
            "open_gopro.api.http_commands",
            "open_gopro.api.ble_commands",
            "open_gopro.domain.communicator_interface",
            "open_gopro.network.ble.adapters.bleak_wrapper",
            "open_gopro.network.ble.client",
            "open_gopro.parsers.bytes",
            "open_gopro.parsers.json",
            "open_gopro.parsers.general",
            "open_gopro.network.wifi.adapters.wireless",
            "open_gopro.network.wifi.mdns_scanner",
            "open_gopro.domain.observable",
            "open_gopro.domain.gopro_observable",
            "open_gopro.models.response",
            "open_gopro.models.network_scan_response",
            "open_gopro.features.cohn_feature",
            "open_gopro.features.access_point_feature",
            "open_gopro.util.util",
            "open_gopro.database.db",
            "bleak",
            "urllib3",
            "http.client",
        ]

        self.logger = logger
        self.modules = modules or self.modules
        self.handlers: list[logging.Handler] = []

        # monkey-patch a `print` global into the http.client module; all calls to
        # print() in that module will then use our logger's debug method
        http_client.HTTPConnection.debuglevel = 1
        http_client.print = lambda *args: logging.getLogger("http.client").debug(" ".join(args))  # type: ignore

        self.file_handler: logging.Handler | None
        if output:
            # Logging to file with millisecond timing
            self.file_handler = logging.FileHandler(output, mode="w")
            file_formatter = logging.Formatter(
                fmt="%(threadName)13s:%(asctime)s.%(msecs)03d %(filename)-40s %(lineno)4s %(levelname)-8s | %(message)s",
                datefmt="%H:%M:%S",
            )
            self.file_handler.setFormatter(file_formatter)
            # Set to TRACE for concurrency debugging
            self.file_handler.setLevel(logging.DEBUG)
            logger.addHandler(self.file_handler)
            self.addLoggingHandler(self.file_handler)
        else:
            self.file_handler = None

        # Use Rich for colorful console logging
        self.stream_handler = RichHandler(rich_tracebacks=True, enable_link_path=True, show_time=False)
        stream_formatter = logging.Formatter("%(asctime)s.%(msecs)03d %(message)s", datefmt="%H:%M:%S")
        self.stream_handler.setFormatter(stream_formatter)
        self.stream_handler.setLevel(logging.INFO)
        logger.addHandler(self.stream_handler)
        self.addLoggingHandler(self.stream_handler)

        self.addLoggingLevel("TRACE", logging.DEBUG - 5)
        logger.setLevel(logging.TRACE)  # type: ignore # pylint: disable=no-member

        traceback.install()  # Enable exception tracebacks in rich logger

    @classmethod
    def get_instance(cls) -> Logger:
        """Get the singleton instance

        Raises:
            RuntimeError: Has not yet been instantiated

        Returns:
            Logger: singleton instance
        """
        if not (logger := cls._instances.get(Logger, None)):
            raise RuntimeError("Logging must first be setup")
        return logger

    def addLoggingHandler(self, handler: logging.Handler) -> None:
        """Add a handler for all of the internal GoPro modules

        Args:
            handler (logging.Handler): handler to add
        """
        self.logger.addHandler(handler)
        self.handlers.append(handler)

        # Enable / disable logging in modules
        for module in self.modules:
            l = logging.getLogger(module)
            l.setLevel(logging.TRACE)  # type: ignore
            l.addHandler(handler)

    # From https://stackoverflow.com/questions/2183233/how-to-add-a-custom-loglevel-to-pythons-logging-facility/35804945#35804945
    @staticmethod
    def addLoggingLevel(levelName: str, levelNum: int) -> None:
        """Comprehensively adds a new logging level to the `logging` module and the currently configured logging class.

        `levelName` becomes an attribute of the `logging` module with the value
        `levelNum`. `methodName` becomes a convenience method for both `logging`
        itself and the class returned by `logging.getLoggerClass()` (usually just
        `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
        used.

        To avoid accidental clobberings of existing attributes, this method will
        raise an `AttributeError` if the level name is already an attribute of the
        `logging` module or if the method name is already present

        Example:
        --------
        >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
        >>> logging.getLogger(__name__).setLevel("TRACE")
        >>> logging.getLogger(__name__).trace('that worked')
        >>> logging.trace('so did this')
        >>> logging.TRACE
        5

        Args:
            levelName (str): name of level (i.e. TRACE)
            levelNum (int): integer level of new logging level
        """
        methodName = levelName.lower()

        def logForLevel(self: Any, message: str, *args: Any, **kwargs: Any) -> None:
            if self.isEnabledFor(levelNum):
                self._log(levelNum, message, args, **kwargs)

        def logToRoot(message: str, *args: Any, **kwargs: Any) -> None:
            logging.log(levelNum, message, *args, **kwargs)

        logging.addLevelName(levelNum, levelName)
        setattr(logging, levelName, levelNum)
        setattr(logging.getLoggerClass(), methodName, logForLevel)
        setattr(logging, methodName, logToRoot)

    @staticmethod
    def build_log_tx_str(stringable: Any) -> str:
        """Build a string with Tx arrows

        Args:
            stringable (Any): stringable object to surround with arrows

        Returns:
            str: string surrounded by Tx arrows
        """
        s = str(stringable).strip(r"{}")
        arrow = f"{'<'*Logger.ARROW_HEAD_COUNT}{'-'*Logger.ARROW_TAIL_COUNT}"
        return f"\n{arrow}{s}{arrow}\n"

    @staticmethod
    def build_log_rx_str(stringable: Any, asynchronous: bool = False) -> str:
        """Build a string with Rx arrows

        Args:
            stringable (Any): stringable object to surround with arrows
            asynchronous (bool): Should the arrows contain ASYNC?. Defaults to False.

        Returns:
            str: string surrounded by Rx arrows
        """
        s = str(stringable).strip(r"{}")
        assert Logger.ARROW_TAIL_COUNT > 5
        if asynchronous:
            arrow = f"{'-'*(Logger.ARROW_TAIL_COUNT//2-3)}ASYNC{'-'*(Logger.ARROW_TAIL_COUNT//2-2)}{'>'*Logger.ARROW_HEAD_COUNT}"
        else:
            arrow = f"{'-'*Logger.ARROW_TAIL_COUNT}{'>'*Logger.ARROW_HEAD_COUNT}"
        return f"\n{arrow}{s}{arrow}\n"


def setup_logging(
    base: logging.Logger | str,
    output: Path | None = None,
    modules: list[str] | None = None,
) -> logging.Logger:
    """Configure the GoPro modules for logging and get a logger that can be used by the application

    This can only be called once and should be done at the top level of the application.

    Args:
        base (logging.Logger | str): Name of application (i.e. __name__) or preconfigured logger to use as base
        output (Path | None): Path of log file for file stream handler. If not set, will not log to file.
        modules (list[str] | None): Optional override of modules / levels. Will be merged into default modules.

    Raises:
        TypeError: Base logger is not of correct type

    Returns:
        logging.Logger: updated logger that the application can use for logging
    """
    if isinstance(base, str):
        base = logging.getLogger(base)
    elif not isinstance(base, logging.Logger):
        raise TypeError("Base must be of type logging.Logger or str")
    l = Logger(base, output, modules)
    return l.logger


def set_file_logging_level(level: int) -> None:
    """Change the global logging level for the default file output handler

    Args:
        level (int): level to set
    """
    if fh := Logger.get_instance().file_handler:
        fh.setLevel(level)


def set_stream_logging_level(level: int) -> None:
    """Change the global logging level for the default stream output handler

    Args:
        level (int): level to set
    """
    Logger.get_instance().stream_handler.setLevel(level)


def set_logging_level(level: int) -> None:
    """Change the global logging level for the default file and stream output handlers

    Args:
        level (int): level to set
    """
    set_file_logging_level(level)
    set_stream_logging_level(level)


def add_logging_handler(handler: logging.Handler) -> None:
    """Add a handler to all of the GoPro internal modules

    Args:
        handler (logging.Handler): handler to add
    """
    Logger.get_instance().addLoggingHandler(handler)



# cohn.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Oct 24 19:08:07 UTC 2023

"""Entrypoint for configuring and demonstrating Camera On the Home Network (COHN)."""

from __future__ import annotations

import argparse
import asyncio
import datetime
import multiprocessing as mp
from dataclasses import dataclass
from pathlib import Path
from threading import Event

from open_gopro import WirelessGoPro
from open_gopro.gopro_wired import WiredGoPro
from open_gopro.models import constants, proto
from open_gopro.models.general import CohnInfo, ScheduledCapture
from open_gopro.util import add_cli_args_and_parse
from open_gopro.util.logger import setup_logging


@dataclass
class GoPro:
    """GoPro Target information"""

    serial: str
    name: str
    gopro: WirelessGoPro | None = None
    cohn: CohnInfo | None = None


def multi_record_via_usb(  # pylint: disable = unused-argument
    target: GoPro,
    record_event: Event,
    ready_event: Event,
    ssid: str | None,
    password: str | None,
) -> None:
    """Multiprocess target to synchronized record via USB

    Args:
        target (GoPro): gopro to communicate with
        record_event (Event): event to wait on to start recording
        ready_event (Event): event to notify manager that this target is ready
        ssid (str | None): not used
        password (str | None): not used
    """
    setup_logging(__name__, Path(f"{target.serial}.log"))

    async def _execute() -> None:
        async with WiredGoPro(target.serial) as gopro:
            await gopro.http_command.load_preset_group(group=proto.EnumPresetGroup.PRESET_GROUP_ID_PHOTO)
            ready_event.set()
            record_event.wait()
            await gopro.http_command.set_shutter(shutter=constants.Toggle.ENABLE)

    asyncio.run(_execute())


def multi_record_via_cohn(
    target: GoPro,
    record_event: Event,
    ready_event: Event,
    ssid: str | None,
    password: str | None,
) -> None:
    """Multiprocess target to synchronized record via COHN

    Args:
        target (GoPro): gopro to communicate with
        record_event (Event): event to wait on to start recording
        ready_event (Event): event to notify manager that this target is ready
        ssid (str | None): WiFi SSID to connect to if COHN provisioning is needed
        password (str | None): WiFi password to use if COHN provisioning is needed
    """
    logger = setup_logging(__name__, Path(f"{target.serial}.log"))

    async def _execute() -> None:
        try:
            # Start with just BLE connected in order to provision COHN
            async with WirelessGoPro(target=target.serial, interfaces={WirelessGoPro.Interface.BLE}) as gopro:
                if await gopro.cohn.is_configured:
                    print("COHN is already configured.")
                else:
                    if not (ssid and password):
                        raise ValueError("COHN provisioning is needed but SSID and password were not provided")
                    await gopro.access_point.connect(ssid, password)
                    await gopro.cohn.configure(force_reprovision=True)
            # # Now use COHN
            async with WirelessGoPro(target=target.serial, interfaces={WirelessGoPro.Interface.COHN}) as gopro:
                await gopro.http_command.load_preset_group(group=proto.EnumPresetGroup.PRESET_GROUP_ID_PHOTO)
                ready_event.set()
                record_event.wait()
                await gopro.http_command.set_shutter(shutter=constants.Toggle.ENABLE)

        except Exception as e:  # pylint: disable = broad-except
            logger.error(repr(e))
            if gopro:
                await gopro.close()

    asyncio.run(_execute())


def multi_record_via_ble(  # pylint: disable = unused-argument
    target: GoPro,
    record_event: Event,
    ready_event: Event,
    ssid: str | None,
    password: str | None,
) -> None:
    """Multiprocess target to synchronized record via BLE

    Args:
        target (GoPro): gopro to communicate with
        record_event (Event): event to wait on to start recording
        ready_event (Event): event to notify manager that this target is ready
        ssid (str | None): not used
        password (str | None): not used
    """
    setup_logging(__name__, Path(f"{target.serial}.log"))

    async def _execute() -> None:
        async with WirelessGoPro(target.serial, interfaces={WirelessGoPro.Interface.BLE}) as gopro:
            await gopro.ble_command.load_preset_group(group=proto.EnumPresetGroup.PRESET_GROUP_ID_PHOTO)
            ready_event.set()
            record_event.wait()
            await gopro.ble_command.set_shutter(shutter=constants.Toggle.ENABLE)

    asyncio.run(_execute())


def multi_record_via_scheduled_capture(  # pylint: disable = unused-argument
    target: GoPro,
    record_event: Event,
    ready_event: Event,
    ssid: str | None,
    password: str | None,
) -> None:
    """Multiprocess target to synchronized record via Scheduled Capture over BLE

    Args:
        target (GoPro): gopro to communicate with
        record_event (Event): event to wait on to start recording
        ready_event (Event): event to notify manager that this target is ready
        ssid (str | None): not used
        password (str | None): not used
    """
    logger = setup_logging(__name__, Path(f"{target.serial}.log"))

    async def _execute() -> None:
        async with WirelessGoPro(target.serial, interfaces={WirelessGoPro.Interface.BLE}) as gopro:
            await gopro.ble_command.load_preset_group(group=proto.EnumPresetGroup.PRESET_GROUP_ID_PHOTO)
            camera_now = (await gopro.ble_command.get_date_time_tz_dst()).data.datetime
            capture_time = camera_now + datetime.timedelta(minutes=1)
            logger.info(f"Camera time is {camera_now}. Recording at {capture_time}")
            await gopro.ble_setting.scheduled_capture.set(
                ScheduledCapture(
                    hour=capture_time.hour,
                    minute=capture_time.minute,
                    is_24_hour=True,
                    is_enabled=True,
                )
            )
            ready_event.set()
            record_event.wait()

    asyncio.run(_execute())


def main(args: argparse.Namespace) -> None:
    gopro_targets = [GoPro(serial, f"Device {idx}") for idx, serial in enumerate(args.devices)]
    record_event = mp.Event()
    ready_events = [mp.Event() for _ in gopro_targets]
    match args.interface:
        case "usb":
            target = multi_record_via_usb
        case "cohn":
            target = multi_record_via_cohn
        case "ble":
            target = multi_record_via_ble
        case "scheduled_capture":
            target = multi_record_via_scheduled_capture
    processes = [
        mp.Process(target=target, args=(gopro_target, record_event, event, args.ssid, args.password))
        for event, gopro_target in zip(ready_events, gopro_targets)
    ]
    # This will serialize the initialization of each process. This is fine since we only care about synchronizing the
    # recording after initialization. And I don't trust OS-level BLE to operate correctly receiving simultaneous
    # commands.
    for event, process in zip(ready_events, processes):
        process.start()
        event.wait()
    record_event.set()
    for process in processes:
        process.join()


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Provision / connect a camera for COHN. SSID and password must be passed if COHN is not currently provisioned."
    )
    parser.add_argument(
        "interface",
        type=str,
        choices=["usb", "cohn", "ble", "scheduled_capture"],
        help="Interface to communicate with the target devices",
    )
    parser.add_argument(
        "devices",
        type=str,
        nargs="+",
        help="list of devices to communicate with as last 4 digits of device serial number",
    )
    parser.add_argument(
        "--ssid",
        type=str,
        help="WiFi SSID to connect to if not currently provisioned for COHN.",
        default=None,
    )
    parser.add_argument(
        "--password",
        type=str,
        help="Password of WiFi SSID.",
        default=None,
    )
    return add_cli_args_and_parse(parser, wifi=False)


# Needed for poetry scripts defined in pyproject.toml
def entrypoint() -> None:
    main(parse_arguments())


if __name__ == "__main__":
    entrypoint()



# settings.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu Feb 20 23:24:52 UTC 2025

"""Setting-related constants"""

########################################################################################################################
#
# Warning!! This file is auto-generated. Do not modify it manually.
#
########################################################################################################################

from open_gopro.domain.enum import GoProIntEnum


class SettingId(GoProIntEnum):
    """Setting ID's that identify settings and are written to GoProUUID.CQ_SETTINGS."""

    VIDEO_RESOLUTION = 2
    FRAMES_PER_SECOND = 3
    VIDEO_TIMELAPSE_RATE = 5
    PHOTO_TIMELAPSE_RATE = 30
    NIGHTLAPSE_RATE = 32
    WEBCAM_DIGITAL_LENSES = 43
    AUTO_POWER_DOWN = 59
    GPS = 83
    LCD_BRIGHTNESS = 88
    LED = 91
    VIDEO_ASPECT_RATIO = 108
    VIDEO_LENS = 121
    PHOTO_LENS = 122
    TIME_LAPSE_DIGITAL_LENSES = 123
    PHOTO_OUTPUT = 125
    MEDIA_FORMAT = 128
    ANTI_FLICKER = 134
    HYPERSMOOTH = 135
    VIDEO_HORIZON_LEVELING = 150
    PHOTO_HORIZON_LEVELING = 151
    VIDEO_DURATION = 156
    MULTI_SHOT_DURATION = 157
    MAX_LENS = 162
    HINDSIGHT = 167
    SCHEDULED_CAPTURE = 168
    PHOTO_SINGLE_INTERVAL = 171
    PHOTO_INTERVAL_DURATION = 172
    VIDEO_PERFORMANCE_MODE = 173
    CONTROL_MODE = 175
    EASY_MODE_SPEED = 176
    ENABLE_NIGHT_PHOTO = 177
    WIRELESS_BAND = 178
    STAR_TRAILS_LENGTH = 179
    SYSTEM_VIDEO_MODE = 180
    VIDEO_BIT_RATE = 182
    BIT_DEPTH = 183
    PROFILES = 184
    VIDEO_EASY_MODE = 186
    LAPSE_MODE = 187
    MAX_LENS_MOD = 189
    MAX_LENS_MOD_ENABLE = 190
    EASY_NIGHT_PHOTO = 191
    MULTI_SHOT_ASPECT_RATIO = 192
    FRAMING = 193
    CAMERA_VOLUME = 216
    SETUP_SCREEN_SAVER = 219
    SETUP_LANGUAGE = 223
    PHOTO_MODE = 227
    VIDEO_FRAMING = 232
    MULTI_SHOT_FRAMING = 233
    FRAME_RATE = 234


class VideoResolution(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#video-resolution-2)"""

    NUM_4K = 1
    NUM_2_7K = 4
    NUM_2_7K_4_3 = 6
    NUM_1440 = 7
    NUM_1080 = 9
    NUM_720 = 12
    NUM_4K_4_3 = 18
    NUM_5K = 24
    NUM_5K_4_3 = 25
    NUM_5_3K_8_7 = 26
    NUM_5_3K_4_3 = 27
    NUM_4K_8_7 = 28
    NUM_5_3K_21_9 = 35
    NUM_4K_21_9 = 36
    NUM_4K_1_1 = 37
    NUM_900 = 38
    NUM_5_3K = 100
    NUM_5_3K_8_7_V2 = 107
    NUM_4K_8_7_V2 = 108
    NUM_4K_9_16_V2 = 109
    NUM_1080_9_16_V2 = 110
    NUM_2_7K_4_3_V2 = 111
    NUM_4K_4_3_V2 = 112
    NUM_5_3K_4_3_V2 = 113


class FramesPerSecond(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#frames-per-second-3)"""

    NUM_240_0 = 0
    NUM_120_0 = 1
    NUM_100_0 = 2
    NUM_60_0 = 5
    NUM_50_0 = 6
    NUM_30_0 = 8
    NUM_25_0 = 9
    NUM_24_0 = 10
    NUM_200_0 = 13
    NUM_400_0 = 15
    NUM_360_0 = 16
    NUM_300_0 = 17


class VideoTimelapseRate(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#video-timelapse-rate-5)"""

    NUM_0_5_SECONDS = 0
    NUM_1_SECOND = 1
    NUM_2_SECONDS = 2
    NUM_5_SECONDS = 3
    NUM_10_SECONDS = 4
    NUM_30_SECONDS = 5
    NUM_60_SECONDS = 6
    NUM_2_MINUTES = 7
    NUM_5_MINUTES = 8
    NUM_30_MINUTES = 9
    NUM_60_MINUTES = 10
    NUM_3_SECONDS = 11


class PhotoTimelapseRate(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#photo-timelapse-rate-30)"""

    NUM_3_SECONDS = 11
    NUM_60_MINUTES = 100
    NUM_30_MINUTES = 101
    NUM_5_MINUTES = 102
    NUM_2_MINUTES = 103
    NUM_60_SECONDS = 104
    NUM_30_SECONDS = 105
    NUM_10_SECONDS = 106
    NUM_5_SECONDS = 107
    NUM_2_SECONDS = 108
    NUM_1_SECOND = 109
    NUM_0_5_SECONDS = 110


class NightlapseRate(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#nightlapse-rate-32)"""

    NUM_4_SECONDS = 4
    NUM_5_SECONDS = 5
    NUM_10_SECONDS = 10
    NUM_15_SECONDS = 15
    NUM_20_SECONDS = 20
    NUM_30_SECONDS = 30
    NUM_60_SECONDS = 100
    NUM_2_MINUTES = 120
    NUM_5_MINUTES = 300
    NUM_30_MINUTES = 1800
    NUM_60_MINUTES = 3600
    AUTO = 3601


class WebcamDigitalLenses(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#webcam-digital-lenses-43)"""

    WIDE = 0
    NARROW = 2
    SUPERVIEW = 3
    LINEAR = 4


class AutoPowerDown(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#auto-power-down-59)"""

    NEVER = 0
    NUM_1_MIN = 1
    NUM_5_MIN = 4
    NUM_15_MIN = 6
    NUM_30_MIN = 7
    NUM_8_SECONDS = 11
    NUM_30_SECONDS = 12


class Gps(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#gps-83)"""

    OFF = 0
    ON = 1


class Led(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#led-91)"""

    OFF = 0
    ON = 2
    ALL_ON = 3
    ALL_OFF = 4
    FRONT_OFF_ONLY = 5
    BACK_ONLY = 100


class VideoAspectRatio(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#video-aspect-ratio-108)"""

    NUM_4_3 = 0
    NUM_16_9 = 1
    NUM_8_7 = 3
    NUM_9_16 = 4
    NUM_21_9 = 5
    NUM_1_1 = 6


class VideoLens(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#video-lens-121)"""

    WIDE = 0
    NARROW = 2
    SUPERVIEW = 3
    LINEAR = 4
    MAX_SUPERVIEW = 7
    LINEAR_HORIZON_LEVELING = 8
    HYPERVIEW = 9
    LINEAR_HORIZON_LOCK = 10
    MAX_HYPERVIEW = 11
    ULTRA_SUPERVIEW = 12
    ULTRA_WIDE = 13
    ULTRA_LINEAR = 14
    ULTRA_HYPERVIEW = 104


class PhotoLens(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#photo-lens-122)"""

    WIDE_12_MP = 0
    LINEAR_12_MP = 10
    NARROW = 19
    WIDE_23_MP = 27
    LINEAR_23_MP = 28
    WIDE_27_MP = 31
    LINEAR_27_MP = 32
    NUM_13MP_LINEAR = 38
    NUM_13MP_WIDE = 39
    NUM_13MP_ULTRA_WIDE = 40
    ULTRA_WIDE_12_MP = 41
    NUM_13MP_ULTRA_LINEAR = 44
    MAX_SUPERVIEW = 100
    WIDE = 101
    LINEAR = 102


class TimeLapseDigitalLenses(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#time-lapse-digital-lenses-123)"""

    NARROW = 19
    WIDE_27_MP = 31
    LINEAR_27_MP = 32
    MAX_SUPERVIEW = 100
    WIDE = 101
    LINEAR = 102


class PhotoOutput(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#photo-output-125)"""

    STANDARD = 0
    RAW = 1
    HDR = 2
    SUPERPHOTO = 3


class MediaFormat(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#media-format-128)"""

    TIME_LAPSE_VIDEO = 13
    TIME_LAPSE_PHOTO = 20
    NIGHT_LAPSE_PHOTO = 21
    NIGHT_LAPSE_VIDEO = 26


class Anti_Flicker(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#anti-flicker-134)"""

    NTSC = 0
    PAL = 1
    NUM_60HZ = 2
    NUM_50HZ = 3


class Hypersmooth(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#hypersmooth-135)"""

    OFF = 0
    LOW = 1
    HIGH = 2
    BOOST = 3
    AUTO_BOOST = 4
    STANDARD = 100


class VideoHorizonLeveling(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#video-horizon-leveling-150)"""

    OFF = 0
    LOCKED = 2


class PhotoHorizonLeveling(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#photo-horizon-leveling-151)"""

    OFF = 0
    LOCKED = 2


class VideoDuration(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#video-duration-156)"""

    NUM_15_SECONDS = 1
    NUM_30_SECONDS = 2
    NUM_1_MINUTE = 3
    NUM_5_MINUTES = 4
    NUM_15_MINUTES = 5
    NUM_30_MINUTES = 6
    NUM_1_HOUR = 7
    NUM_2_HOURS = 8
    NUM_3_HOURS = 9
    NUM_5_SECONDS = 10
    NO_LIMIT = 100


class MultiShotDuration(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#multi-shot-duration-157)"""

    NUM_15_SECONDS = 1
    NUM_30_SECONDS = 2
    NUM_1_MINUTE = 3
    NUM_5_MINUTES = 4
    NUM_15_MINUTES = 5
    NUM_30_MINUTES = 6
    NUM_1_HOUR = 7
    NUM_2_HOURS = 8
    NUM_3_HOURS = 9
    NO_LIMIT = 100


class MaxLens(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#max-lens-162)"""

    OFF = 0
    ON = 1


class Hindsight(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#hindsight-167)"""

    NUM_15_SECONDS = 2
    NUM_30_SECONDS = 3
    OFF = 4


class PhotoSingleInterval(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#photo-single-interval-171)"""

    OFF = 0
    NUM_0_5S = 2
    NUM_1S = 3
    NUM_2S = 4
    NUM_5S = 5
    NUM_10S = 6
    NUM_30S = 7
    NUM_60S = 8
    NUM_120S = 9
    NUM_3S = 10


class PhotoIntervalDuration(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#photo-interval-duration-172)"""

    OFF = 0
    NUM_15_SECONDS = 1
    NUM_30_SECONDS = 2
    NUM_1_MINUTE = 3
    NUM_5_MINUTES = 4
    NUM_15_MINUTES = 5
    NUM_30_MINUTES = 6
    NUM_1_HOUR = 7
    NUM_2_HOURS = 8
    NUM_3_HOURS = 9


class VideoPerformanceMode(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#video-performance-mode-173)"""

    MAXIMUM_VIDEO_PERFORMANCE = 0
    EXTENDED_BATTERY = 1
    TRIPOD_STATIONARY_VIDEO = 2


class ControlMode(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#control-mode-175)"""

    EASY = 0
    PRO = 1


class EasyModeSpeed(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#easy-mode-speed-176)"""

    NUM_8X_ULTRA_SLO_MO = 0
    NUM_4X_SUPER_SLO_MO = 1
    NUM_2X_SLO_MO = 2
    NUM_1X_SPEED_LOW_LIGHT_ = 3
    NUM_4X_SUPER_SLO_MO_EXT_BATT_ = 4
    NUM_2X_SLO_MO_EXT_BATT_ = 5
    NUM_1X_SPEED_EXT_BATT_LOW_LIGHT_ = 6
    NUM_8X_ULTRA_SLO_MO_50HZ_ = 7
    NUM_4X_SUPER_SLO_MO_50HZ_ = 8
    NUM_2X_SLO_MO_50HZ_ = 9
    NUM_1X_SPEED_50HZ_LOW_LIGHT_ = 10
    NUM_4X_SUPER_SLO_MO_50HZ_EXT_BATT_ = 11
    NUM_2X_SLO_MO_50HZ_EXT_BATT_ = 12
    NUM_1X_SPEED_50HZ_EXT_BATT_LOW_LIGHT_ = 13
    NUM_8X_ULTRA_SLO_MO_EXT_BATT_ = 14
    NUM_8X_ULTRA_SLO_MO_50HZ_EXT_BATT_ = 15
    NUM_8X_ULTRA_SLO_MO_LONG_BATT_ = 16
    NUM_4X_SUPER_SLO_MO_LONG_BATT_ = 17
    NUM_2X_SLO_MO_LONG_BATT_ = 18
    NUM_1X_SPEED_LONG_BATT_LOW_LIGHT_ = 19
    NUM_8X_ULTRA_SLO_MO_50HZ_LONG_BATT_ = 20
    NUM_4X_SUPER_SLO_MO_50HZ_LONG_BATT_ = 21
    NUM_2X_SLO_MO_50HZ_LONG_BATT_ = 22
    NUM_1X_SPEED_50HZ_LONG_BATT_LOW_LIGHT_ = 23
    NUM_2X_SLO_MO_4K_ = 24
    NUM_4X_SUPER_SLO_MO_2_7K_ = 25
    NUM_2X_SLO_MO_4K_50HZ_ = 26
    NUM_4X_SUPER_SLO_MO_2_7K_50HZ_ = 27
    NUM_8X_ULTRA_SLO_MO_V2_ = 100
    NUM_4X_SUPER_SLO_MO_V2_ = 101
    NUM_2X_SLO_MO_V2_ = 102
    NUM_1X_SPEED_LOW_LIGHT_V2_ = 103
    NUM_8X_ULTRA_SLO_MO_50HZ_V2_ = 104
    NUM_4X_SUPER_SLO_MO_50HZ_V2_ = 105
    NUM_2X_SLO_MO_50HZ_V2_ = 106
    NUM_1X_SPEED_50HZ_LOW_LIGHT_V2_ = 107
    NUM_8X_ULTRA_SLO_MO_LONG_BATT_V2_ = 108
    NUM_4X_SUPER_SLO_MO_LONG_BATT_V2_ = 109
    NUM_2X_SLO_MO_LONG_BATT_V2_ = 110
    NUM_1X_SPEED_LONG_BATT_LOW_LIGHT_V2_ = 111
    NUM_8X_ULTRA_SLO_MO_50HZ_LONG_BATT_V2_ = 112
    NUM_4X_SUPER_SLO_MO_50HZ_LONG_BATT_V2_ = 113
    NUM_2X_SLO_MO_50HZ_LONG_BATT_V2_ = 114
    NUM_1X_SPEED_50HZ_LONG_BATT_LOW_LIGHT_V2_ = 115
    NUM_2X_SLO_MO_4K_V2_ = 116
    NUM_2X_SLO_MO_4K_50HZ_V2_ = 117
    NUM_1X_SPEED_LOW_LIGHT_V2_VERTICAL_ = 118
    NUM_1X_SPEED_50HZ_LOW_LIGHT_V2_VERTICAL_ = 119
    NUM_2X_SLO_MO_V2_VERTICAL_ = 120
    NUM_2X_SLO_MO_50HZ_V2_VERTICAL_ = 121
    NUM_1X_SPEED_FULL_FRAME_LOW_LIGHT_V2_ = 122
    NUM_1X_SPEED_50HZ_FULL_FRAME_LOW_LIGHT_V2_ = 123
    NUM_2X_SLO_MO_FULL_FRAME_V2_ = 124
    NUM_2X_SLO_MO_50HZ_FULL_FRAME_V2_ = 125
    NUM_1X_SPEED_4K_LOW_LIGHT_V2_ = 126
    NUM_1X_SPEED_4K_50HZ_LOW_LIGHT_V2_ = 127
    NUM_1X_SPEED_2_7K_LOW_LIGHT_V2_ = 128
    NUM_1X_SPEED_2_7K_50HZ_LOW_LIGHT_V2_ = 129
    NUM_2X_SLO_MO_2_7K_V2_ = 130
    NUM_2X_SLO_MO_2_7K_50HZ_V2_ = 131
    NUM_2X_SLO_MO_LONG_BATT_V2_VERTICAL_ = 132
    NUM_2X_SLO_MO_50HZ_LONG_BATT_V2_VERTICAL_ = 133
    NUM_1X_SPEED_LONG_BATT_LOW_LIGHT_V2_VERTICAL_ = 134
    NUM_1X_SPEED_50HZ_LONG_BATT_LOW_LIGHT_V2_VERTICAL_ = 135
    NUM_1X_SPEED_4K_FULL_FRAME_LOW_LIGHT_V2_ = 136
    NUM_1X_SPEED_4K_50HZ_FULL_FRAME_LOW_LIGHT_V2_ = 137
    NUM_1X_NORMAL_SPEED_1_1_30_FPS_4K_V2_ = 138
    NUM_1X_NORMAL_SPEED_1_1_25_FPS_4K_V2_ = 139
    NUM_2X_SLO_MO_SPEED_1_1_4K_60_FPS_V2_ = 140
    NUM_2X_SLO_MO_SPEED_1_1_4K_50_FPS_V2_ = 141
    NUM_1X_NORMAL_SPEED_21_9_30_FPS_5_3K_V2_ = 142
    NUM_1X_NORMAL_SPEED_21_9_25_FPS_5_3K_V2_ = 143
    NUM_2X_SLO_MO_SPEED_21_9_5_3K_60_FPS_V2_ = 144
    NUM_2X_SLO_MO_SPEED_21_9_5_3K_50_FPS_V2_ = 145
    NUM_1X_NORMAL_SPEED_21_9_30_FPS_4K_V2_ = 146
    NUM_1X_NORMAL_SPEED_21_9_25_FPS_4K_V2_ = 147
    NUM_2X_SLO_MO_SPEED_21_9_4K_60_FPS_V2_ = 148
    NUM_2X_SLO_MO_SPEED_21_9_4K_50_FPS_V2_ = 149
    NUM_120_4X_SUPER_SLO_MO_SPEED_21_9_4K_V2_ = 150
    NUM_100_4X_SUPER_SLO_MO_SPEED_21_9_4K_V2_ = 151
    NUM_1X_NORMAL_SPEED_30_FPS_4_3_5_3K_V2_ = 152
    NUM_1X_NORMAL_SPEED_25_FPS_4_3_5_3K_V2_ = 153
    NUM_1X_NORMAL_SPEED_30_FPS_4_3_4K_V2_ = 154
    NUM_1X_NORMAL_SPEED_25_FPS_4_3_4K_V2_ = 155
    NUM_2X_SLO_MO_SPEED_4_3_4K_60_FPS_V2_ = 156
    NUM_2X_SLO_MO_SPEED_4_3_4K_50_FPS_V2_ = 157
    NUM_120_4X_SUPER_SLO_MO_SPEED_2_7K_4_3_V2_ = 158
    NUM_100_4X_SUPER_SLO_MO_SPEED_2_7K_4_3_V2_ = 159


class EnableNightPhoto(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#enable-night-photo-177)"""

    OFF = 0
    ON = 1


class WirelessBand(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#wireless-band-178)"""

    NUM_2_4GHZ = 0
    NUM_5GHZ = 1


class StarTrailsLength(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#star-trails-length-179)"""

    SHORT = 1
    LONG = 2
    MAX = 3


class SystemVideoMode(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#system-video-mode-180)"""

    HIGHEST_QUALITY = 0
    EXTENDED_BATTERY = 101
    LONGEST_BATTERY = 102
    STANDARD_QUALITY = 111
    BASIC_QUALITY = 112


class VideoBitRate(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#video-bit-rate-182)"""

    STANDARD = 0
    HIGH = 1


class BitDepth(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#bit-depth-183)"""

    NUM_8_BIT = 0
    NUM_10_BIT = 2


class Profiles(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#profiles-184)"""

    STANDARD = 0
    HDR = 1
    LOG = 2
    HLG_HDR = 101


class VideoEasyMode(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#video-easy-mode-186)"""

    HIGHEST_QUALITY = 0
    STANDARD_QUALITY = 1
    BASIC_QUALITY = 2
    STANDARD_VIDEO = 3
    HDR_VIDEO = 4


class LapseMode(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#lapse-mode-187)"""

    TIMEWARP = 0
    STAR_TRAILS = 1
    LIGHT_PAINTING = 2
    VEHICLE_LIGHTS = 3
    MAX_TIMEWARP = 4
    MAX_STAR_TRAILS = 5
    MAX_LIGHT_PAINTING = 6
    MAX_VEHICLE_LIGHTS = 7
    TIME_LAPSE_VIDEO = 8
    NIGHT_LAPSE_VIDEO = 9


class MaxLensMod(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#max-lens-mod-189)"""

    NONE = 0
    MAX_LENS_1_0 = 1
    MAX_LENS_2_0 = 2
    MAX_LENS_2_5 = 3
    MACRO = 4
    ANAMORPHIC = 5
    ND_4 = 6
    ND_8 = 7
    ND_16 = 8
    ND_32 = 9
    STANDARD_LENS = 10
    AUTO_DETECT = 100


class MaxLensModEnable(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#max-lens-mod-enable-190)"""

    OFF = 0
    ON = 1


class EasyNightPhoto(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#easy-night-photo-191)"""

    SUPER_PHOTO = 0
    NIGHT_PHOTO = 1
    BURST = 2


class MultiShotAspectRatio(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#multi-shot-aspect-ratio-192)"""

    NUM_4_3 = 0
    NUM_16_9 = 1
    NUM_8_7 = 3
    NUM_9_16 = 4


class Framing(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#framing-193)"""

    WIDESCREEN = 0
    VERTICAL = 1
    FULL_FRAME = 2
    TRADITIONAL_4_3_V2 = 100
    WIDESCREEN_16_9_V2 = 101
    FULL_FRAME_8_7_V2 = 103
    VERTICAL_9_16_V2 = 104
    ULTRA_WIDESCREEN_21_9_V2 = 105
    FULL_FRAME_1_1_V2 = 106


class CameraVolume(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#camera-volume-216)"""

    LOW = 70
    MEDIUM = 85
    HIGH = 100


class SetupScreenSaver(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#setup-screen-saver-219)"""

    NUM_1_MIN = 1
    NUM_2_MIN = 2
    NUM_3_MIN = 3
    NUM_5_MIN = 4


class SetupLanguage(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#setup-language-223)"""

    ENGLISH_US = 0
    ENGLISH_UK = 1
    ENGLISH_AUS = 2
    GERMAN = 3
    FRENCH = 4
    ITALIAN = 5
    SPANISH = 6
    SPANISH_NA = 7
    CHINESE = 8
    JAPANESE = 9
    KOREAN = 10
    PORTUGUESE = 11
    RUSSIAN = 12
    ENGLISH_IND = 13
    SWEDISH = 14


class PhotoMode(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#photo-mode-227)"""

    SUPERPHOTO = 0
    NIGHT_PHOTO = 1
    BURST = 2


class VideoFraming(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#video-framing-232)"""

    NUM_4_3 = 0
    NUM_16_9 = 1
    NUM_8_7 = 3
    NUM_9_16 = 4
    NUM_21_9 = 5
    NUM_1_1 = 6


class MultiShotFraming(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#multi-shot-framing-233)"""

    NUM_4_3 = 0
    NUM_16_9 = 1
    NUM_8_7 = 3
    NUM_9_16 = 4


class FrameRate(GoProIntEnum):
    """See [Open GoPro Spec](https://gopro.github.io/OpenGoPro/ble/features/settings.html#frame-rate-234)"""

    NUM_240_0 = 0
    NUM_120_0 = 1
    NUM_100_0 = 2
    NUM_60_0 = 5
    NUM_50_0 = 6
    NUM_30_0 = 8
    NUM_25_0 = 9
    NUM_24_0 = 10
    NUM_200_0 = 13
    NUM_400_0 = 15
    NUM_360_0 = 16
    NUM_300_0 = 17



# wifi_controller.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:51 PM

"""Manage a WiFI connection using native OS commands."""

from __future__ import annotations

import asyncio
import ctypes
import html
import locale
import logging
import os
import platform
import re
import tempfile
import time
from enum import Enum, auto
from getpass import getpass
from shutil import which
from typing import Any, Callable, Optional

import wrapt
from packaging.version import Version

from open_gopro.network.wifi import SsidState, WifiController
from open_gopro.util import cmd

logger = logging.getLogger(__name__)


def ensure_us_english() -> None:
    """Validate the system language is US English for CLI response parsing

    From https://stackoverflow.com/questions/3425294/how-to-detect-the-os-default-language-in-python

    Raises:
        RuntimeError: The system is using any language other then en_US
    """
    if platform.system().lower() == "windows":
        windll = getattr(ctypes, "windll").kernel32
        language = locale.windows_locale[windll.GetUserDefaultUILanguage()]
    else:
        language = os.environ["LANG"]

    if not language.startswith("en_US"):
        raise RuntimeError(
            f"The Wifi driver parses CLI responses and only supports en_US where your language is {language}"
        )


@wrapt.decorator
def pass_through_to_driver(wrapped: Callable, instance: WifiCli, args: Any, kwargs: Any) -> Any:
    """Call this same method on the _driver attribute

    Args:
        wrapped (Callable): method to call
        instance (WifiCli): instance to use to find driver
        args (Any): positional arguments
        kwargs (Any): keyword arguments

    Returns:
        Any: result after pass through to driver
    """
    driver_method = getattr(instance._driver, wrapped.__name__)
    return driver_method(*args, **kwargs)


class WifiCli(WifiController):
    """Top level abstraction of different Wifi drivers.

    If interface is not specified (i.e. it is None), we will attempt to automatically
    discover a suitable interface

    This will raise a RunTimeError if either:
        - The system is using any language other then en_US
        - We weren't able to find a suitable driver or auto-detect an interface after detecting driver

    Args:
        interface (str | None): Interface. Defaults to None.
        password (str | None): User Password for sudo. Defaults to None.
    """

    def __init__(self, interface: str | None = None, password: str | None = None) -> None:  # noqa: DOC502
        ensure_us_english()
        WifiController.__init__(self, interface, password)

        # detect and init appropriate driver
        self._driver = self._detect_driver()

        # Attempt to set interface (will raise an exception if not able to auto-detect)
        self.interface = interface  # type: ignore

        logger.debug(f"Wifi setup. Using {self}")

    def __str__(self) -> str:
        return f"[{type(self).__name__}] driver::[{self.interface}] interface"

    def _detect_driver(self) -> WifiController:
        """Try to find and instantiate a Wifi driver that can be used.

        Raises:
            RuntimeError: We weren't able to find a suitable driver or auto-detect an interface after detecting driver

        Returns:
            WifiController: [description]
        """
        # Try netsh (Windows).
        if os.name == "nt" and which("netsh"):
            return NetshWireless()

        # try networksetup (Mac OS 10.10)
        if which("networksetup"):
            return NetworksetupWireless()

        # Try Linux options.
        # try nmcli (Ubuntu 14.04). Allow for use in Snap Package
        if which("nmcli") or which("nmcli", path="/snap/bin/"):

            ctrl_wifi = cmd("nmcli general permissions |grep enable-disable-wifi")
            scan_wifi = cmd("nmcli general permissions |grep scan")

            if not "yes" in ctrl_wifi or not "yes" in scan_wifi:
                self._sudo_from_stdin()

            version = cmd("nmcli --version").split()[-1]
            # On RHEL based systems, the version is in the form of 1.44.2-1.fc39
            # wich raises an error when trying to compare it with the Version class
            if any(c.isalpha() for c in version):
                version = version.split("-")[0]
            return (
                Nmcli0990Wireless(password=self._password)
                if Version(version) >= Version("0.9.9.0")
                else NmcliWireless(password=self._password)
            )
        # try nmcli (Ubuntu w/o network-manager)
        if which("wpa_supplicant"):
            self._sudo_from_stdin()
            return WpasupplicantWireless(password=self._password)

        raise RuntimeError("Unable to find compatible wireless driver.")

    def _sudo_from_stdin(self) -> str:
        """Ask for sudo password input from stdin

        This method prompts the user to enter the sudo password from the command line.
        It validates the password by running a command with sudo and checking if the password is valid.

        Returns:
            str: The entered sudo password.

        Raises:
            RuntimeError: If the password is empty or invalid.
        """
        # Need password for sudo
        if not self._password:
            self._password = getpass("Need to run as sudo. Enter password: ")
        if not self._password:
            raise RuntimeError("Can't use sudo with empty password.")
        # Validate password
        if "VALID PASSWORD" not in cmd(f'echo "{self._password}" | sudo -S echo "VALID PASSWORD"'):
            raise RuntimeError("Invalid password")

        return self._password

    @pass_through_to_driver
    async def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:  # type: ignore
        """Connect to a network.

        Args:
            ssid (str): SSID of network to connect to
            password (str): password of network to connect to
            timeout (float): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: True if successful, False otherwise
        """

    @pass_through_to_driver
    async def disconnect(self) -> bool:  # type: ignore
        """Disconnect from a network.

        Returns:
            bool: True if successful, False otherwise
        """

    @pass_through_to_driver
    def current(self) -> tuple[Optional[str], SsidState]:  # type: ignore
        """Return the SSID and state of the current network.

        Returns:
            tuple[Optional[str], SsidState]: Tuple of SSID str and state. If SSID is None,
            there is no current connection.
        """

    @pass_through_to_driver
    def available_interfaces(self) -> list[str]:  # type: ignore
        """Return a list of the available Wifi interfaces

        Returns:
            list[str]: list of available interfaces
        """

    @pass_through_to_driver
    def power(self, power: bool) -> bool:  # type: ignore
        """Enable / disable the wireless driver.

        Args:
            power (bool): Enable if True. Disable if False.

        Returns:
            bool: True if successful, False otherwise
        """

    @property
    def interface(self) -> str:
        """Get the Wifi Interface

        Returns:
            str: interface
        """
        return self._driver.interface

    @interface.setter
    def interface(self, interface: Optional[str]) -> None:
        """Set the Wifi interface.

        If None is passed, interface will attempt to be auto-detected

        Args:
            interface (Optional[str]): interface (or None)
        """
        self._driver.interface = interface  # type: ignore

    @property
    def is_on(self) -> bool:
        """Is the wireless driver currently enabled.

        Returns:
            bool: True if yes. False if no.
        """
        return self._driver.is_on


class NmcliWireless(WifiController):
    """Linux nmcli Driver < 0.9.9.0."""

    def __init__(self, password: str | None, interface: Optional[str] = None) -> None:
        WifiController.__init__(self, interface=interface, password=password)

    def _clean(self, partial: str) -> None:
        """Clean up connections.

        This is needed to prevent the following error after extended use:
        'maximum number of pending replies per connection has been reached'

        Args:
            partial (str): part of the connection name
        """
        # list matching connections
        response = cmd(f'{self.sudo} nmcli --fields BleUUID,NAME con list | grep "{partial}"')

        # delete all of the matching connections
        for line in response.splitlines():
            if len(line) > 0:
                cmd(f"{self.sudo} nmcli con delete uuid {line.split()[0]}")

    @staticmethod
    def _error_in_response(response: str) -> bool:
        """Ignore warnings in nmcli output.

        Sometimes there are warnings but we connected just fine

        Args:
            response (str): output to parse

        Returns:
            bool: True if errors found. False if not.
        """
        # no error if no response
        if len(response) == 0:
            return False

        # loop through each line
        for line in response.splitlines():
            # all error lines start with 'Error'
            if line.startswith("Error"):
                return True

        # if we didn't find an error then we are in the clear
        return False

    async def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """Connect to WiFi SSID.

        Args:
            ssid (str): network SSID
            password (str): network password
            timeout (float): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: [description]
        """

        def _sync_connect() -> bool:
            nonlocal ssid
            nonlocal password
            nonlocal timeout
            # clean up previous connection
            current, _ = self.current()
            if current is not None:
                self._clean(current)

            # First scan to ensure our network is there
            logger.info(f"Scanning for {ssid}...")
            cmd(f"{self.sudo} nmcli device wifi rescan")

            start = time.time()
            discovered = False
            while not discovered and (time.time() - start) <= timeout:
                # Scan for network
                response = cmd(f"{self.sudo} nmcli -f SSID device wifi list")
                for result in response.splitlines()[1:]:  # Skip title row
                    if result.strip() == ssid.strip():
                        discovered = True
                        break
                if discovered:
                    break
                time.sleep(1)
            else:
                logger.warning("Wifi Scan timed out")
                return False

            # attempt to connect
            logger.info(f"Connecting to {ssid}...")
            response = cmd(
                f'{self.sudo} nmcli dev wifi connect "{ssid}" password "{password}" iface "{self.interface}"'
            )

            # parse response
            return not self._error_in_response(response)

        return await asyncio.to_thread(_sync_connect)

    async def disconnect(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return False

    def current(self) -> tuple[Optional[str], SsidState]:
        """[summary].

        Returns:
            tuple[Optional[str], SsidState]: [description]
        """
        # list active connections for all interfaces
        response = cmd(f'{self.sudo} nmcli con status | grep "{self.interface}"')

        # the current network is in the first column
        for line in response.splitlines():
            if len(line) > 0:
                return (line.split()[0], SsidState.CONNECTED)

        # return none if there was not an active connection
        return (None, SsidState.DISCONNECTED)

    def available_interfaces(self) -> list[str]:
        """Return a list of available Wifi Interface strings

        Returns:
            list[str]: list of interfaces
        """
        # grab list of interfaces
        response = cmd(f"{self.sudo} nmcli dev")

        # parse response
        interfaces = []
        for line in response.splitlines():
            if "wireless" in line:
                # this line has our interface name in the first column
                interfaces.append(line.split()[0])

        # return list
        return interfaces

    @property
    def is_on(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return "enabled" in cmd(f"{self.sudo} nmcli nm wifi")

    def power(self, power: bool) -> bool:
        """Enable or disbale the Wifi controller

        Args:
            power (bool): True to enable, False to Disable

        Returns:
            bool: True if success, False otherwise
        """
        if power:
            cmd(f"{self.sudo} nmcli nm wifi on")
        else:
            cmd(f"{self.sudo} nmcli nm wifi off")

        return True


class Nmcli0990Wireless(WifiController):
    """Linux nmcli Driver >= 0.9.9.0."""

    def __init__(self, password: str | None, interface: Optional[str] = None) -> None:
        WifiController.__init__(self, interface=interface, password=password)

    def _clean(self, partial: str) -> None:
        """Clean up connections.

        This is needed to prevent the following error after extended use:
        'maximum number of pending replies per connection has been reached'

        Args:
            partial (str): part of the connection name
        """
        # list matching connections
        response = cmd(f'{self.sudo} nmcli --fields UUID,NAME con show | grep "{partial}"')

        # delete all of the matching connections
        for line in response.splitlines():
            if len(line) > 0:
                uuid = line.split()[0]
                cmd(f"{self.sudo} nmcli con delete uuid {uuid}")

    @staticmethod
    def _error_in_response(response: str) -> bool:
        """Ignore warnings in nmcli output.

        Sometimes there are warnings but we connected just fine

        Args:
            response (str): output to parse

        Returns:
            bool: True if errors found. False if not.
        """
        # no error if no response
        if len(response) == 0:
            return False

        # loop through each line
        for line in response.splitlines():
            # all error lines start with 'Error'
            if line.startswith("Error"):
                logger.error("line")
                return True

        # if we didn't find an error then we are in the clear
        return False

    async def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """Connect to Wifi SSID.

        Args:
            ssid (str): network SSID
            password (str): network password
            timeout (float): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: True if connection suceeded, False otherwise
        """

        def _sync_connect() -> bool:
            nonlocal ssid
            nonlocal password
            nonlocal timeout
            # First scan to ensure our network is there
            logger.info(f"Scanning for {ssid}...")
            start = time.time()
            discovered = False
            while not discovered and (time.time() - start) <= timeout:
                # Scan for network
                cmd(f"{self.sudo} nmcli device wifi rescan")
                response = cmd(f"{self.sudo} nmcli -f SSID device wifi list")
                for result in response.splitlines()[1:]:  # Skip title row
                    if result.strip() == ssid.strip():
                        discovered = True
                        break
                if discovered:
                    break
                time.sleep(1)
            else:
                logger.warning("Wifi Scan timed out")
                return False

            # attempt to connect
            logger.info(f"Connecting to {ssid}...")
            response = cmd(
                f'{self.sudo} nmcli dev wifi connect "{ssid}" password "{password}" ifname "{self.interface}"'
            )

            # parse response
            return not self._error_in_response(response)

        return await asyncio.to_thread(_sync_connect)

    async def disconnect(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return False

    def current(self) -> tuple[Optional[str], SsidState]:
        """[summary].

        Returns:
            tuple[Optional[str], SsidState]: [description]
        """
        # list active connections for all interfaces
        response = cmd(f'{self.sudo} nmcli con | grep "{self.interface}"')

        # the current network is in the first column
        for line in response.splitlines():
            if len(line) > 0:
                return (line.split()[0], SsidState.CONNECTED)

        # return none if there was not an active connection
        return (None, SsidState.DISCONNECTED)

    def available_interfaces(self) -> list[str]:
        """Return a list of available Wifi Interface strings

        Returns:
            list[str]: list of interfaces
        """
        # grab list of interfaces
        response = cmd(f"{self.sudo} nmcli dev")

        # parse response
        interfaces = []
        for line in response.splitlines():
            if "wifi" in line:
                # this line has our interface name in the first column
                interfaces.append(line.split()[0])

        # return list
        return interfaces

    @property
    def is_on(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return "enabled" in cmd(f"{self.sudo} nmcli r wifi")

    def power(self, power: bool) -> bool:
        """Enable or disbale the Wifi controller

        Args:
            power (bool): True to enable, False to Disable

        Returns:
            bool: True if success, False otherwise
        """
        if power:
            cmd(f"{self.sudo} nmcli r wifi on")
        else:
            cmd(f"{self.sudo} nmcli r wifi off")

        return True


class WpasupplicantWireless(WifiController):
    """Linux wpa_supplicant Driver."""

    _file = "/tmp/wpa_supplicant.conf"

    def __init__(self, password: str | None, interface: Optional[str] = None) -> None:
        WifiController.__init__(self, interface=interface, password=password)

    async def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """[summary].

        Args:
            ssid (str): network SSID
            password (str): network password
            timeout (float): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: [description]
        """

        def _sync_connect() -> bool:
            nonlocal ssid
            nonlocal password
            nonlocal timeout
            # attempt to stop any active wpa_supplicant instances
            # ideally we do this just for the interface we care about
            cmd(f"{self.sudo} killall wpa_supplicant")

            # don't do DHCP for GoPros; can cause dropouts with the server
            cmd(f'{self.sudo} ifconfig "{self.interface}" 10.5.5.10/24 up')

            # create configuration file
            with open(self._file, "w") as fp:
                fp.write(f'network={{\n    ssid="{ssid}"\n    psk="{password}"\n}}\n')
                fp.close()

            # attempt to connect
            cmd(f'{self.sudo} wpa_supplicant -i"{self.interface}" -c"{self._file}" -B')

            # check that the connection was successful
            # i've never seen it take more than 3 seconds for the link to establish
            time.sleep(5)
            current_ssid, _ = self.current()
            if current_ssid != ssid:
                return False

            # attempt to grab an IP
            # better hope we are connected because the timeout here is really long
            # cmd(f"{self.sudo} dhclient {self.interface}"")

            # parse response
            return True

        return await asyncio.to_thread(_sync_connect)

    async def disconnect(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return False

    def current(self) -> tuple[Optional[str], SsidState]:
        """[summary].

        Returns:
            tuple[Optional[str], SsidState]: [description]
        """
        # get interface status
        response = cmd(f'{self.sudo} iwconfig "{self.interface}"')

        # the current network is on the first line like ESSID:"network"
        line = response.splitlines()[0]
        line = line.replace('"', "")
        parts = line.split("ESSID:")
        if len(parts) > 1:
            network = parts[1].strip()
            if network != "off/any":
                return (network, SsidState.CONNECTED)

        # return none if there was not an active connection
        return (None, SsidState.DISCONNECTED)

    def available_interfaces(self) -> list[str]:
        """Return a list of available Wifi Interface strings

        Returns:
            list[str]: list of interfaces
        """
        # grab list of interfaces
        response = cmd(f"{self.sudo} iwconfig")

        # parse response
        interfaces = []
        for line in response.splitlines():
            if len(line) > 0 and not line.startswith(" "):
                # this line contains an interface name!
                if "no wireless extensions" not in line:
                    # this is a wireless interface
                    interfaces.append(line.split()[0])
        return interfaces

    @property
    def is_on(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return "enabled" in cmd(f"{self.sudo} nmcli r wifi")

    def power(self, power: bool) -> bool:
        """Enable or disable the Wifi controller

        Args:
            power (bool): True to enable, False to Disable

        Returns:
            bool: True if success, False otherwise
        """
        return False


class NetworksetupWireless(WifiController):
    """OS X networksetup Driver."""

    def __init__(self, interface: Optional[str] = None) -> None:
        WifiController.__init__(self, interface)

    async def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """Connect to SSID.

        Args:
            ssid (str): network SSID
            password (str): network password
            timeout (float): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: [description]
        """

        def _sync_connect() -> bool:
            nonlocal ssid
            nonlocal password
            nonlocal timeout
            # Escape single quotes
            ssid = ssid.replace(r"'", '''"'"''')

            logger.info(f"Scanning for {ssid}...")
            start = time.time()
            discovered = False
            while not discovered and (time.time() - start) <= timeout:
                # Scan for network
                response = cmd(r"/usr/sbin/system_profiler SPAirPortDataType")
                regex = re.compile(
                    r"\n\s+([\x20-\x7E]{1,32}):\n\s+PHY Mode:"
                )  # 0x20...0x7E --> ASCII for printable characters
                if ssid in sorted(regex.findall(response)):
                    break
                time.sleep(1)
            else:
                logger.warning("Wifi Scan timed out")
                return False

            # If we're already connected, return
            if self.current()[0] == ssid:
                return True

            # Connect now that we found the ssid
            logger.info(f"Connecting to {ssid}...")
            response = cmd(f"networksetup -setairportnetwork '{self.interface}' '{ssid}' '{password}'")

            if "not find" in response.lower():
                return False
            # Now wait for network to actually establish
            current = self.current()[0]
            logger.debug(f"current wifi: {current}")
            while current is not None and ssid not in current and timeout > 0:
                time.sleep(1)
                current = self.current()[0]
                logger.debug(f"current wifi: {current}")
                timeout -= 1
                if timeout == 0:
                    return False

            # There is some delay required here, presumably because the network is not ready.
            time.sleep(5)

            return True

        return await asyncio.to_thread(_sync_connect)

    async def disconnect(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return False

    def current(self) -> tuple[Optional[str], SsidState]:
        """[summary].

        Returns:
            tuple[Optional[str], SsidState]: [description]
        """
        # attempt to get current network
        response = cmd(f"networksetup -getairportnetwork '{self.interface}'")

        # parse response
        phrase = "Current Wi-Fi Network: "
        if phrase in response:
            return (response.replace("Current Wi-Fi Network: ", "").strip(), SsidState.CONNECTED)
        return (None, SsidState.DISCONNECTED)

    def available_interfaces(self) -> list[str]:
        """Return a list of available Wifi Interface strings

        Returns:
            list[str]: list of interfaces
        """
        # grab list of interfaces
        response = cmd("networksetup -listallhardwareports")

        # parse response
        interfaces = []
        detected_wifi = False
        for line in response.splitlines():
            if detected_wifi:
                # this line has our interface name in it
                interfaces.append(line.replace("Device: ", ""))
                detected_wifi = False
            else:
                # search for the line that has 'Wi-Fi' in it
                if "Wi-Fi" in line:
                    detected_wifi = True

        # return list
        return interfaces

    @property
    def is_on(self) -> bool:
        """[summary].

        Returns:
            bool: [description]
        """
        return "On" in cmd(f"networksetup -getairportpower '{self.interface}'")

    def power(self, power: bool) -> bool:
        """Enable or disbale the Wifi controller

        Args:
            power (bool): True to enable, False to Disable

        Returns:
            bool: True if success, False otherwise
        """
        cmd(f"networksetup -setairportpower '{self.interface}' {'on' if power else 'off'}")

        return True


class NetshWireless(WifiController):
    """Windows Driver."""

    # Used to build profile
    template = r"""<?xml version="1.0"?>
<WLANProfile xmlns="http://www.microsoft.com/networking/WLAN/profile/v1">
    <name>{ssid}</name>
    <SSIDConfig>
        <SSID>
            <name>{ssid}</name>
        </SSID>
    </SSIDConfig>
    <connectionType>ESS</connectionType>
    <connectionMode>manual</connectionMode>
    <MSM>
        <security>
            <authEncryption>
                <authentication>{auth}</authentication>
                <encryption>{encrypt}</encryption>
                <useOneX>false</useOneX>
            </authEncryption>
            <sharedKey>
                <keyType>passPhrase</keyType>
                <protected>false</protected>
                <keyMaterial>{passwd}</keyMaterial>
            </sharedKey>
        </security>
    </MSM>
    <MacRandomization xmlns="http://www.microsoft.com/networking/WLAN/profile/v3">
        <enableRandomization>false</enableRandomization>
    </MacRandomization>
</WLANProfile>"""

    def __init__(self, interface: Optional[str] = None) -> None:
        WifiController.__init__(self, interface)
        self.ssid: Optional[str] = None

    def __del__(self) -> None:
        self._clean(self.ssid)

    async def connect(self, ssid: str, password: str, timeout: float = 15) -> bool:
        """Establish a connection.

        This is blocking and won't return until either a connection is established or
        a 10 second timeout

        Raises a RuntimeError if it can not add profile or request to connect to SSID fails

        Args:
            ssid (str): SSID of network to connect to
            password (str): password of network to connect to
            timeout (float): Time before considering connection failed (in seconds). Defaults to 15.

        Returns:
            bool: True if connected, False otherwise
        """

        def _sync_connect() -> bool:
            nonlocal password
            nonlocal ssid
            nonlocal timeout
            # Replace xml tokens (&, <, >, etc.)
            password = html.escape(password)
            ssid = html.escape(ssid)

            logger.info(f"Attempting to establish Wifi connection to {ssid}...")

            # Start fresh each time.
            self._clean(ssid)

            # Create new profile
            output = NetshWireless.template.format(ssid=ssid, auth="WPA2PSK", encrypt="AES", passwd=password)
            logger.debug(f"Using template {output}")
            # Need ugly low level mkstemp and os here because standard tempfile can't be accessed by a subprocess in Windows :(
            fd, filename = tempfile.mkstemp()
            os.write(fd, output.encode("utf-8"))
            os.close(fd)
            response = cmd(f"netsh wlan add profile filename={filename}")
            if "is added on interface" not in response:
                raise RuntimeError(response)
            os.remove(filename)

            for _ in range(5):
                # Try to connect
                response = cmd(f'netsh wlan connect ssid="{ssid}" name="{ssid}" interface="{self.interface}"')
                if "was completed successfully" not in response:
                    raise RuntimeError(response)
                # Wait for connection to establish
                DELAY = 1
                while current := self.current():
                    if current == (ssid, SsidState.CONNECTED):
                        logger.info("Wifi connection established!")
                        self.ssid = ssid
                        return True
                    logger.debug(f"Waiting {DELAY} second for Wi-Fi connection to establish...")
                    time.sleep(DELAY)
                    timeout -= DELAY
                    if timeout <= 0 or current[1] is SsidState.DISCONNECTED:
                        logger.debug("Wifi driver detected disconnect. Attempting retry...")
                        break

            return False

        return await asyncio.to_thread(_sync_connect)

    async def disconnect(self) -> bool:
        """Terminate the Wifi connection.

        Returns:
            bool: True if the disconnect was successful, False otherwise.
        """
        response = cmd(f'netsh wlan disconnect interface="{self.interface}"')

        return bool("completed successfully" in response.lower())

    def current(self) -> tuple[Optional[str], SsidState]:
        """Get the current network SSID and state.

        # Here is an example of what we are parsing (i.e. to find FunHouse SSID):
        # Name                   : Wi-Fi
        # Description            : TP-Link Wireless USB Adapter
        # GUID                   : 093d8022-33cb-4400-8362-275eaf24cb86
        # Physical address       : 98:48:27:88:cb:18
        # State                  : connected
        # SSID                   : FunHouse

        Returns:
            tuple[Optional[str], SsidState]: Tuple of (ssid, network_state)
        """

        class ParseState(Enum):
            """Current state of interface parsing"""

            PARSE_INTERFACE = auto()
            PARSE_SSID = auto()
            PARSE_STATE = auto()

        response = cmd("netsh wlan show interfaces")
        parse_state = ParseState.PARSE_INTERFACE
        ssid: Optional[str] = None
        network_state: Optional[str] = None
        for field in response.split("\r\n"):
            if parse_state is ParseState.PARSE_INTERFACE:
                if "Name" in field and self.interface in field:
                    parse_state = ParseState.PARSE_STATE
            elif parse_state is ParseState.PARSE_STATE:
                if "State" in field:
                    network_state = field.split(":")[1].strip().lower()
                    parse_state = ParseState.PARSE_SSID
            elif parse_state is ParseState.PARSE_SSID:
                if "SSID" in field:
                    ssid = field.split(":")[1].strip()
                    break

        if network_state == "connected":
            state = SsidState.CONNECTED
        elif network_state == "disconnected":
            state = SsidState.DISCONNECTED
        else:
            state = SsidState.ESTABLISHING
        return (ssid, state)

    def available_interfaces(self) -> list[str]:
        """Discover all available interfaces.

        # We're parsing, for example, the following line to find "Wi-Fi":
        # Name                   : Wi-Fi

        Returns:
            list[str]: List of interfaces
        """
        response = cmd("netsh wlan show interfaces")
        interfaces = []

        # Look behind to find field, then match (non-greedy) any chars until CRLF
        match = "(?<={}).+?(?=\\r\\n)"
        for interface in re.findall(match.format("Name"), response):
            # Strip leading whitespace and then the first two chars of remaining (i.e. " :")
            interfaces.append(interface.strip()[2:])

        return interfaces

    @property
    def is_on(self) -> bool:
        """Is Wifi enabled?

        For Windows, this means "Is there at least one Wifi interfaces available?"

        Returns:
            bool: True if yes, False if no.
        """
        response = cmd("netsh wlan show interfaces")
        # Is there at least one interfaces enabled?
        if "no wireless interface" in response.lower():
            return False
        return True

    def power(self, power: bool) -> bool:
        """Enable or disable the Wifi controller

        Args:
            power (bool): True to enable, False to Disable

        Returns:
            bool: True if success, False otherwise
        """
        arg = "enable" if power else "disable"
        response = cmd(f'netsh interface set interface "{self._interface}" "{arg}"')
        return "not exist" not in response

    @staticmethod
    def _clean(ssid: Optional[str]) -> None:
        """Disconnect and delete SSID profile.

        Args:
            ssid (Optional[str]): name of SSID
        """
        cmd("netsh wlan disconnect")
        if ssid is not None:
            cmd(f'netsh wlan delete profile name="{ssid}"')



# set_camera_control_status_pb2.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu Apr 24 17:26:18 UTC 2025

"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_sym_db = _symbol_database.Default()
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1fset_camera_control_status.proto\x12\nopen_gopro"c\n\x1dRequestSetCameraControlStatus\x12B\n\x15camera_control_status\x18\x01 \x02(\x0e2#.open_gopro.EnumCameraControlStatus*q\n\x17EnumCameraControlStatus\x12\x0f\n\x0bCAMERA_IDLE\x10\x00\x12\x12\n\x0eCAMERA_CONTROL\x10\x01\x12\x1b\n\x17CAMERA_EXTERNAL_CONTROL\x10\x02\x12\x14\n\x10CAMERA_COF_SETUP\x10\x03'
)
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "set_camera_control_status_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _ENUMCAMERACONTROLSTATUS._serialized_start = 148
    _ENUMCAMERACONTROLSTATUS._serialized_end = 261
    _REQUESTSETCAMERACONTROLSTATUS._serialized_start = 47
    _REQUESTSETCAMERACONTROLSTATUS._serialized_end = 146



# __init__.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Mon Apr 21 22:24:00 UTC 2025

"""Parser interface and implementations"""

from .general import IntByteParserBuilder, ScheduledCaptureParser



# uuids.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu Feb 20 23:24:52 UTC 2025

"""UUID-related constants"""

from dataclasses import dataclass
from typing import Final

from open_gopro.network.ble import BleUUID, UUIDs

GOPRO_BASE_UUID: Final = "b5f9{}-aa8d-11e3-9046-0002a5d5c51b"


@dataclass(frozen=True)
class GoProUUID(UUIDs):
    """GoPro Proprietary BleUUID's."""

    # GoPro Wifi Access Point Service
    S_WIFI_ACCESS_POINT = BleUUID("Wifi Access Point Service", hex=GOPRO_BASE_UUID.format("0001"))
    WAP_SSID = BleUUID("Wifi AP SSID", hex=GOPRO_BASE_UUID.format("0002"))
    WAP_PASSWORD = BleUUID("Wifi AP Password", hex=GOPRO_BASE_UUID.format("0003"))
    WAP_POWER = BleUUID("Wifi Power", hex=GOPRO_BASE_UUID.format("0004"))
    WAP_STATE = BleUUID("Wifi State", hex=GOPRO_BASE_UUID.format("0005"))
    WAP_CSI_PASSWORD = BleUUID("CSI Password", hex=GOPRO_BASE_UUID.format("0006"))

    # GoPro Control & Query Service
    S_CONTROL_QUERY = BleUUID("Control and Query Service", hex="0000fea6-0000-1000-8000-00805f9b34fb")
    CQ_COMMAND = BleUUID("Command", hex=GOPRO_BASE_UUID.format("0072"))
    CQ_COMMAND_RESP = BleUUID("Command Response", hex=GOPRO_BASE_UUID.format("0073"))
    CQ_SETTINGS = BleUUID("Settings", hex=GOPRO_BASE_UUID.format("0074"))
    CQ_SETTINGS_RESP = BleUUID("Settings Response", hex=GOPRO_BASE_UUID.format("0075"))
    CQ_QUERY = BleUUID("Query", hex=GOPRO_BASE_UUID.format("0076"))
    CQ_QUERY_RESP = BleUUID("Query Response", hex=GOPRO_BASE_UUID.format("0077"))
    CQ_SENSOR = BleUUID("Sensor", hex=GOPRO_BASE_UUID.format("0078"))
    CQ_SENSOR_RESP = BleUUID("Sensor Response", hex=GOPRO_BASE_UUID.format("0079"))

    # GoPro Camera Management Service
    S_CAMERA_MANAGEMENT = BleUUID("Camera Management Service", hex=GOPRO_BASE_UUID.format("0090"))
    CM_NET_MGMT_COMM = BleUUID("Camera Management", hex=GOPRO_BASE_UUID.format("0091"))
    CN_NET_MGMT_RESP = BleUUID("Camera Management Response", hex=GOPRO_BASE_UUID.format("0092"))

    # Unknown
    S_INTERNAL = BleUUID("Unknown Service", hex=GOPRO_BASE_UUID.format("0080"))
    INTERNAL_81 = BleUUID("Internal 81", hex=GOPRO_BASE_UUID.format("0081"))
    INTERNAL_82 = BleUUID("Internal 82", hex=GOPRO_BASE_UUID.format("0082"))
    INTERNAL_83 = BleUUID("Internal 83", hex=GOPRO_BASE_UUID.format("0083"))
    INTERNAL_84 = BleUUID("Internal 84", hex=GOPRO_BASE_UUID.format("0084"))



# media_pb2.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu Apr 24 17:26:18 UTC 2025

"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_sym_db = _symbol_database.Default()
from . import response_generic_pb2 as response__generic__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0bmedia.proto\x12\nopen_gopro\x1a\x16response_generic.proto"\x1d\n\x1bRequestGetLastCapturedMedia"l\n\x19ResponseLastCapturedMedia\x12-\n\x06result\x18\x01 \x01(\x0e2\x1d.open_gopro.EnumResultGeneric\x12 \n\x05media\x18\x02 \x01(\x0b2\x11.open_gopro.Media'
)
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "media_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _REQUESTGETLASTCAPTUREDMEDIA._serialized_start = 51
    _REQUESTGETLASTCAPTUREDMEDIA._serialized_end = 80
    _RESPONSELASTCAPTUREDMEDIA._serialized_start = 82
    _RESPONSELASTCAPTUREDMEDIA._serialized_end = 190



# preview_stream.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Mon Jul 31 17:04:07 UTC 2023

"""Example to start and view a preview stream"""

import argparse
import asyncio

from rich.console import Console

from open_gopro import WirelessGoPro
from open_gopro.demos.gui.util import display_video_blocking
from open_gopro.models import constants
from open_gopro.util import add_cli_args_and_parse
from open_gopro.util.logger import setup_logging

console = Console()


async def main(args: argparse.Namespace) -> None:
    setup_logging(__name__, args.log)

    async with WirelessGoPro(args.identifier) as gopro:
        await gopro.http_command.set_preview_stream(mode=constants.Toggle.DISABLE)
        await gopro.ble_command.set_shutter(shutter=constants.Toggle.DISABLE)
        assert (await gopro.http_command.set_preview_stream(mode=constants.Toggle.ENABLE, port=args.port)).ok

        console.print("Displaying the preview stream...")
        display_video_blocking(f"udp://127.0.0.1:{args.port}", printer=console.print)

        await gopro.http_command.set_preview_stream(mode=constants.Toggle.DISABLE)


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Connect to the GoPro via BLE and Wifi, start a preview stream, then display it with CV2."
    )
    parser.add_argument(
        "--port", type=int, help="Port to use for livestream. Defaults to 8554 if not set", default=8554
    )
    return add_cli_args_and_parse(parser, wifi=False)


def entrypoint() -> None:
    asyncio.run(main(parse_arguments()))


if __name__ == "__main__":
    entrypoint()



# __init__.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu Feb 23 22:43:52 UTC 2023

"""Common GUI demos. Ensure that additional GUI options were installed before continuing"""

import sys

try:
    import cv2
    import PIL

except ModuleNotFoundError:
    print(
        "Required GUI dependencies not found. See the installation steps: https://gopro.github.io/OpenGoPro/python_sdk/installation.html"
    )
    sys.exit(1)



# api.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:52 UTC 2021

"""Implementation of Open GoPro API version 2.0"""

from __future__ import annotations

from typing import Final

from open_gopro.domain.communicator_interface import GoProHttp, GoProWirelessInterface

from .ble_commands import BleAsyncResponses, BleCommands
from .ble_settings import BleSettings
from .ble_statuses import BleStatuses
from .http_commands import HttpCommands
from .http_settings import HttpSettings


class WirelessApi:
    """Implementation of Open GoPro API version 2.0 for Wireless interface (Wifi and BLE)

    Attributes:
        version (Final[str]): The API version that this object implements

    Args:
        communicator (GoProWirelessInterface): used to communicate via BLE and Wifi
    """

    version: Final[str] = "2.0"

    def __init__(self, communicator: GoProWirelessInterface) -> None:
        self._communicator = communicator
        self.ble_command = BleCommands(communicator)
        self.ble_setting = BleSettings(communicator)
        self.ble_status = BleStatuses(communicator)
        BleAsyncResponses.add_parsers()
        self.http_command = HttpCommands(communicator)
        self.http_setting = HttpSettings(communicator)


class WiredApi:
    """Implementation of Open GoPro API version 2.0 for Wired interface (USB)

    Attributes:
        version (Final[str]): The API version that this object implements

    Args:
        communicator (GoProHttp): used to communicate via BLE and Wifi
    """

    version: Final[str] = "2.0"

    def __init__(self, communicator: GoProHttp) -> None:
        self._communicator = communicator
        self.http_command = HttpCommands(communicator)
        self.http_setting = HttpSettings(communicator)



# parsers.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Mon Jun 26 18:26:05 UTC 2023

"""Parser Protocol and Bases"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Any, Callable, ClassVar, Generic, Protocol, TypeVar, cast

from open_gopro.domain.types import JsonDict, ResponseType
from open_gopro.models.constants import ActionId, FeatureId

logger = logging.getLogger(__name__)

T_co = TypeVar("T_co", covariant=True)
T = TypeVar("T")

########################################################################################
####### Transformers
########################################################################################


class BaseTransformer(ABC, Generic[T]):
    """Transformer interface.

    A transformer is something that transforms the input into the same output type
    """

    @abstractmethod
    def transform(self, data: T) -> T:
        """Transform data into output matching the input type

        Args:
            data (T): data to transform

        Returns:
            T: transformed data
        """


class BytesTransformer(BaseTransformer[bytes]):
    """Bytes to Bytes transformer interface"""


class JsonTransformer(BaseTransformer[JsonDict]):
    """Json to json transformer interface"""


########################################################################################
####### Parsers
########################################################################################


class BaseParser(ABC, Generic[T, T_co]):
    """Base Parser Interface

    A parser is something that transforms input into a different type
    """

    @abstractmethod
    def parse(self, data: T) -> T_co:  # pylint: disable=method-hidden
        """Parse data into output type

        Args:
            data (T): input data to parse

        Returns:
            T_co: parsed output
        """


class JsonParser(BaseParser[JsonDict, T_co]):
    """Json to Target Type Parser Interface"""


class BytesParser(BaseParser[bytes, T_co]):
    """Bytes to Target Type Parser Interface"""


class Parser(ABC, Generic[T]):
    """The common monolithic Parser that is used for all byte and json parsing / transforming

    Algorithm is:
        1. Variable number of byte transformers (bytes --> bytes)
        2. One bytes Json adapter (bytes --> json)
        3. Variable number of json transformers (json --> json)
        4. One JSON parser (json -> T)

    Args:
        byte_transformers (list[BytesTransformer] | None): bytes --> bytes. Defaults to None.
        byte_json_adapter (BytesParser[JsonDict] | None): bytes --> json. Defaults to None.
        json_transformers (list[JsonTransformer] | None): json --> json. Defaults to None.
        json_parser (JsonParser[T] | None): json --> T. Defaults to None.
    """

    def __init__(
        self,
        byte_transformers: list[BytesTransformer] | None = None,
        byte_json_adapter: BytesParser[JsonDict] | None = None,
        json_transformers: list[JsonTransformer] | None = None,
        json_parser: JsonParser[T] | None = None,
    ) -> None:
        self.byte_transformers = byte_transformers or []
        self.byte_json_adapter = byte_json_adapter
        self.json_transformers = json_transformers or []
        self.json_parser = json_parser

    def parse(self, data: bytes | bytearray | JsonDict) -> T:
        """Perform the parsing using the stored transformers and parsers

        Args:
            data (bytes | bytearray | JsonDict): input bytes or json to parse

        Raises:
            RuntimeError: attempted to parse bytes when a byte-json adapter does not exist

        Returns:
            T: final parsed output
        """
        parsed_json: JsonDict
        if isinstance(data, (bytes, bytearray)):
            data = bytes(data)
            if not self.byte_json_adapter:
                raise RuntimeError("Can not parse bytes without Json Adapter")
            # Filter bytes
            parsed_bytes = bytes(data)
            for byte_transformer in self.byte_transformers:
                parsed_bytes = byte_transformer.transform(data)
            parsed_json = self.byte_json_adapter.parse(parsed_bytes)
        else:
            parsed_json = data

        for json_transformer in self.json_transformers:
            parsed_json = json_transformer.transform(parsed_json)
        if self.json_parser:
            return self.json_parser.parse(parsed_json)
        return cast(T, parsed_json)


########################################################################################
####### Builders
########################################################################################


class BytesBuilder(Protocol):
    """Base bytes serializer protocol definition"""

    def build(self, obj: Any) -> bytes:
        """Build bytestream from object

        Args:
            obj (Any): object to serialize

        Returns:
            bytes: serialized bytestream
        """


class BytesParserBuilder(BytesParser[T_co], BytesBuilder):
    """Class capable of both building / parsing bytes to / from object"""

    @abstractmethod
    def parse(self, data: bytes) -> T_co:
        """Parse input bytes to output

        Args:
            data (bytes): data to parsed

        Returns:
            T_co: parsed output
        """

    @abstractmethod
    def build(self, obj: Any) -> bytes:
        """Build bytestream from object

        Args:
            obj (Any): object to serialize

        Returns:
            bytes: serialized bytestream
        """


class GlobalParsers:
    """Parsers that relate globally to ID's as opposed to contextualized per-message

    This is intended to be used as a singleton, i.e. not instantiated
    """

    _feature_action_id_map: ClassVar[dict[FeatureId, list[ActionId]]] = defaultdict(list)
    _global_parsers: ClassVar[dict[ResponseType, Parser]] = {}

    @classmethod
    def add_feature_action_id_mapping(cls, feature_id: FeatureId, action_id: ActionId) -> None:
        """Add a feature id-to-action id mapping entry

        Args:
            feature_id (FeatureId): Feature ID of protobuf command
            action_id (ActionId): Action ID of protobuf command
        """
        cls._feature_action_id_map[feature_id].append(action_id)

    @classmethod
    def add(cls, identifier: ResponseType, parser: Parser) -> None:
        """Add a global parser that can be accessed by this class's class methods

        Args:
            identifier (ResponseType): identifier to add parser for
            parser (Parser): parser to add
        """
        cls._global_parsers[identifier] = parser

    @classmethod
    def get_query_container(cls, identifier: ResponseType) -> Callable | None:
        """Attempt to get a callable that will translate an input value to the ID-appropriate value.

        For example, _get_query_container(SettingId.VIDEO_RESOLUTION) will return
        :py:class:`open_gopro.models.constants.settings.VideoResolution`

        As another example, _get_query_container(StatusId.TURBO_TRANSFER) will return bool()

        Note! Not all ID's are currently parsed so None will be returned if the container does not exist

        Args:
            identifier (ResponseType): identifier to find container for

        Returns:
            Callable | None: container if found else None
        """
        try:
            parser_builder = cast(BytesParserBuilder, cls._global_parsers[identifier].byte_json_adapter)
            return lambda data, parse=parser_builder.parse, build=parser_builder.build: parse(build(data))
        except KeyError:
            return None

    @classmethod
    def get_parser(cls, identifier: ResponseType) -> Parser | None:
        """Get a globally defined parser for the given ID.

        Currently, only BLE uses globally defined parsers

        Args:
            identifier (ResponseType): ID to get parser for

        Returns:
            Parser | None: parser if found, else None
        """
        return cls._global_parsers.get(identifier)



# gopro.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:47 PM

"""Implements top level interface to GoPro module."""

from __future__ import annotations

import asyncio
import enum
import json
import logging
import threading
import traceback
from abc import abstractmethod
from typing import Any, Awaitable, Callable, Final, Generic, TypeVar

import requests
import wrapt

from open_gopro.api import (
    BleCommands,
    BleSettings,
    BleStatuses,
    HttpCommands,
    HttpSettings,
    WiredApi,
    WirelessApi,
)
from open_gopro.domain.communicator_interface import (
    GoProHttp,
    HttpMessage,
    Message,
    MessageRules,
)
from open_gopro.domain.exceptions import GoProNotOpened, ResponseTimeout
from open_gopro.domain.types import JsonDict
from open_gopro.models import GoProResp
from open_gopro.models.constants import ErrorCode
from open_gopro.parsers.response import RequestsHttpRespBuilderDirector
from open_gopro.util import pretty_print
from open_gopro.util.logger import Logger

logger = logging.getLogger(__name__)

GoPro = TypeVar("GoPro", bound="GoProBase")
ApiType = TypeVar("ApiType", WiredApi, WirelessApi)
MessageMethodType = Callable[[Any, bool], Awaitable[GoProResp]]


class GoProMessageInterface(enum.Enum):
    """Enum to identify wireless interface"""

    HTTP = enum.auto()
    BLE = enum.auto()


@wrapt.decorator
def catch_thread_exception(wrapped: Callable, instance: GoProBase, args: Any, kwargs: Any) -> Callable | None:
    """Catch any exceptions from this method and pass them to the exception handler identifier by thread name

    Args:
        wrapped (Callable): method that this is wrapping
        instance (GoProBase): instance owner of method
        args (Any): positional args
        kwargs (Any): keyword args

    Returns:
        Callable | None: forwarded return of wrapped method or None if exception occurs
    """
    try:
        return wrapped(*args, **kwargs)
    except Exception as e:  # pylint: disable=broad-exception-caught
        instance._handle_exception(threading.current_thread().name, {"exception": e})
        return None


def ensure_opened(interface: tuple[GoProMessageInterface]) -> Callable:
    """Raise exception if relevant interface is not currently opened

    Args:
        interface (tuple[GoProMessageInterface]): wireless interface to verify

    Returns:
        Callable: Direct pass-through of callable after verification
    """

    @wrapt.decorator
    def wrapper(wrapped: Callable, instance: GoProBase, args: Any, kwargs: Any) -> Callable:
        if GoProMessageInterface.BLE in interface and not instance.is_ble_connected:
            raise GoProNotOpened("BLE not connected")
        if GoProMessageInterface.HTTP in interface and not instance.is_http_connected:
            raise GoProNotOpened("HTTP interface not connected")
        return wrapped(*args, **kwargs)

    return wrapper


@wrapt.decorator
async def enforce_message_rules(wrapped: MessageMethodType, instance: GoProBase, args: Any, kwargs: Any) -> GoProResp:
    """Decorator proxy to call the GoProBase's _enforce_message_rules method.

    Args:
        wrapped (MessageMethodType): Operation to enforce
        instance (GoProBase): GoProBase instance to use
        args (Any): positional arguments to wrapped
        kwargs (Any): keyword arguments to wrapped

    Returns:
        GoProResp: common response object
    """
    return await instance._enforce_message_rules(wrapped, *args, **kwargs)


class GoProBase(GoProHttp, Generic[ApiType]):
    """The base class for communicating with all GoPro Clients"""

    HTTP_TIMEOUT: Final = 5
    HTTP_GET_RETRIES: Final = 5

    def __init__(self, **kwargs: Any) -> None:
        self._should_maintain_state = kwargs.get("maintain_state", True)
        self._exception_cb = kwargs.get("exception_cb", None)

    async def __aenter__(self: GoPro) -> GoPro:
        await self.open()
        return self

    async def __aexit__(self, *_: Any) -> None:
        await self.close()

    @abstractmethod
    async def open(self, timeout: int = 10, retries: int = 5) -> None:
        """Connect to the GoPro Client and prepare it for communication

        Args:
            timeout (int): time before considering connection a failure. Defaults to 10.
            retries (int): number of connection retries. Defaults to 5.
        """

    @abstractmethod
    async def close(self) -> None:
        """Gracefully close the GoPro Client connection"""

    @property
    @abstractmethod
    async def is_ready(self) -> bool:
        """Is gopro ready to receive commands

        Returns:
            bool: yes if ready, no otherwise
        """

    @property
    @abstractmethod
    def _requests_session(self) -> requests.Session:
        """The requests session used to communicate with the GoPro Client

        Returns:
            requests.Session: requests session
        """

    @property
    @abstractmethod
    def identifier(self) -> str:
        """Unique identifier for the connected GoPro Client

        Returns:
            str: identifier
        """

    @property
    def version(self) -> str:
        """The API version that the connected camera supports

        Only 2.0 is currently supported

        Returns:
            str: supported version
        """
        return self._api.version

    @property
    @abstractmethod
    def http_command(self) -> HttpCommands:
        """Used to access the Wifi commands

        Returns:
            HttpCommands: the commands
        """

    @property
    @abstractmethod
    def http_setting(self) -> HttpSettings:
        """Used to access the Wifi settings

        Returns:
            HttpSettings: the settings
        """

    @property
    @abstractmethod
    def ble_command(self) -> BleCommands:
        """Used to call the BLE commands

        Returns:
            BleCommands: the commands
        """

    @property
    @abstractmethod
    def ble_setting(self) -> BleSettings:
        """Used to access the BLE settings

        Returns:
            BleSettings: the settings
        """

    @property
    @abstractmethod
    def ble_status(self) -> BleStatuses:
        """Used to access the BLE statuses

        Returns:
            BleStatuses: the statuses
        """

    @property
    @abstractmethod
    def is_open(self) -> bool:
        """Is this client ready for communication?

        Returns:
            bool: True if yes, False if no
        """

    @property
    @abstractmethod
    def is_ble_connected(self) -> bool:
        """Are we connected via BLE to the GoPro device?

        Returns:
            bool: True if yes, False if no
        """

    @property
    @abstractmethod
    def is_http_connected(self) -> bool:
        """Are we connected via HTTP to the GoPro device?

        Returns:
            bool: True if yes, False if no
        """

    ##########################################################################################################
    #                                 End Public API
    ##########################################################################################################

    @abstractmethod
    async def _enforce_message_rules(
        self, wrapped: Callable, message: Message, rules: MessageRules, **kwargs: Any
    ) -> GoProResp:
        """Rule Enforcer. Called by enforce_message_rules decorator.

        Args:
            wrapped (Callable): operation to enforce
            message (Message): message passed to operation
            rules (MessageRules): rules to enforce
            **kwargs (Any) : arguments passed to operation

        Returns:
            GoProResp: Operation response
        """

    def _handle_exception(self, source: Any, context: JsonDict) -> None:
        """Gather exceptions from module threads and send through callback if registered.

        Note that this function signature matches asyncio's exception callback requirement.

        Args:
            source (Any): Where did the exception come from?
            context (JsonDict): Access exception via context["exception"]
        """
        # context["message"] will always be there; but context["exception"] may not
        if exception := context.get("exception", False):
            logger.error(f"Received exception {exception} from {source}")
            logger.error(traceback.format_exc())
            if self._exception_cb:
                self._exception_cb(exception)
        else:
            logger.error(f"Caught unknown message: {context['message']} from {source}")

    class _InternalState(enum.IntFlag):
        """State used to manage whether the GoPro instance is ready or not."""

        READY = 0
        ENCODING = 1 << 0
        SYSTEM_BUSY = 1 << 1

    @property
    @abstractmethod
    def _base_url(self) -> str:
        """Build the base endpoint for USB commands

        Returns:
            str: base endpoint with URL from serial number
        """

    @property
    @abstractmethod
    def _api(self) -> ApiType:
        """Unique identifier for the connected GoPro Client

        Returns:
            ApiType: identifier
        """

    @staticmethod
    def _ensure_opened(interface: tuple[GoProMessageInterface]) -> Callable:
        """Raise exception if relevant interface is not currently opened

        Args:
            interface (tuple[GoProMessageInterface]): wireless interface to verify

        Returns:
            Callable: Direct pass-through of callable after verification
        """
        return ensure_opened(interface)

    @staticmethod
    def _catch_thread_exception(*args: Any, **kwargs: Any) -> Callable | None:
        """Catch any exceptions from this method and pass them to the exception handler identifier by thread name

        Args:
            *args (Any): positional args
            **kwargs (Any): keyword args

        Returns:
            Callable | None: forwarded return of wrapped method or None if exception occurs
        """
        return catch_thread_exception(*args, **kwargs)

    def _build_http_request_args(self, message: HttpMessage) -> dict[str, Any]:
        """Helper method to build request kwargs from message

        Args:
            message (HttpMessage): message to build args from

        Returns:
            dict[str, Any]: built args
        """
        # Dynamically build get kwargs
        request_args: dict[str, Any] = {}
        if message._headers:
            request_args["headers"] = message._headers
        if message._certificate:
            request_args["verify"] = str(message._certificate)
        return request_args

    @enforce_message_rules
    async def _get_json(
        self,
        message: HttpMessage,
        *,
        timeout: int = HTTP_TIMEOUT,
        rules: MessageRules = MessageRules(),
        **kwargs: Any,
    ) -> GoProResp:
        url = self._base_url + message.build_url(**kwargs)
        logger.debug(f"Sending:  {url}")
        logger.info(Logger.build_log_tx_str(pretty_print(message._as_dict(**kwargs))))
        for retry in range(1, GoProBase.HTTP_GET_RETRIES + 1):
            try:
                http_response = self._requests_session.get(
                    url,
                    timeout=timeout,
                    **self._build_http_request_args(message),
                )
                logger.trace(f"received raw json: {json.dumps(http_response.json() if http_response.text else {}, indent=4)}")  # type: ignore
                if not http_response.ok:
                    logger.warning(f"Received non-success status {http_response.status_code}: {http_response.reason}")
                response = RequestsHttpRespBuilderDirector(http_response, message._parser)()
                break
            except requests.exceptions.ConnectionError as e:
                # This appears to only occur after initial connection after pairing
                logger.warning(repr(e))
                # Back off before retrying. TODO This appears to be needed on MacOS
                await asyncio.sleep(2)
            except Exception as e:  # pylint: disable=broad-exception-caught
                logger.critical(f"Unexpected error: {repr(e)}")
            logger.warning(f"Retrying #{retry} to send the command...")
        else:
            raise ResponseTimeout(GoProBase.HTTP_GET_RETRIES)

        logger.info(Logger.build_log_rx_str(pretty_print(response._as_dict())))
        return response

    @enforce_message_rules
    async def _get_stream(
        self,
        message: HttpMessage,
        *,
        timeout: int = HTTP_TIMEOUT,
        rules: MessageRules = MessageRules(),
        **kwargs: Any,
    ) -> GoProResp:
        url = self._base_url + message.build_url(path=kwargs["camera_file"])
        logger.debug(f"Sending:  {url}")
        with self._requests_session.get(
            url,
            stream=True,
            timeout=timeout,
            **self._build_http_request_args(message),
        ) as request:
            request.raise_for_status()
            file = kwargs["local_file"]
            with open(file, "wb") as f:
                logger.debug(f"receiving stream to {file}...")
                for chunk in request.iter_content(chunk_size=8192):
                    f.write(chunk)

        return GoProResp(protocol=GoProResp.Protocol.HTTP, status=ErrorCode.SUCCESS, data=file, identifier=url)

    @enforce_message_rules
    async def _put_json(
        self,
        message: HttpMessage,
        *,
        timeout: int = HTTP_TIMEOUT,
        rules: MessageRules = MessageRules(),
        **kwargs: Any,
    ) -> GoProResp:
        url = self._base_url + message.build_url(**kwargs)
        body = message.build_body(**kwargs)
        logger.debug(f"Sending:  {url} with body: {json.dumps(body, indent=4)}")
        for retry in range(1, GoProBase.HTTP_GET_RETRIES + 1):
            try:
                http_response = self._requests_session.put(
                    url,
                    timeout=timeout,
                    json=body,
                    **self._build_http_request_args(message),
                )
                logger.trace(f"received raw json: {json.dumps(http_response.json() if http_response.text else {}, indent=4)}")  # type: ignore
                if not http_response.ok:
                    logger.warning(f"Received non-success status {http_response.status_code}: {http_response.reason}")
                response = RequestsHttpRespBuilderDirector(http_response, message._parser)()
                break
            except requests.exceptions.ConnectionError as e:
                # This appears to only occur after initial connection after pairing
                logger.warning(repr(e))
                # Back off before retrying. TODO This appears to be needed on MacOS
                await asyncio.sleep(2)
            except Exception as e:  # pylint: disable=broad-exception-caught
                logger.critical(f"Unexpected error: {repr(e)}")
            logger.warning(f"Retrying #{retry} to send the command...")
        else:
            raise ResponseTimeout(GoProBase.HTTP_GET_RETRIES)

        return response



# cohn_pb2.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu Apr 24 17:26:18 UTC 2025

"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_sym_db = _symbol_database.Default()
from . import response_generic_pb2 as response__generic__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\ncohn.proto\x12\nopen_gopro\x1a\x16response_generic.proto"4\n\x14RequestGetCOHNStatus\x12\x1c\n\x14register_cohn_status\x18\x01 \x01(\x08"\xd9\x01\n\x10NotifyCOHNStatus\x12*\n\x06status\x18\x01 \x01(\x0e2\x1a.open_gopro.EnumCOHNStatus\x12/\n\x05state\x18\x02 \x01(\x0e2 .open_gopro.EnumCOHNNetworkState\x12\x10\n\x08username\x18\x03 \x01(\t\x12\x10\n\x08password\x18\x04 \x01(\t\x12\x11\n\tipaddress\x18\x05 \x01(\t\x12\x0f\n\x07enabled\x18\x06 \x01(\x08\x12\x0c\n\x04ssid\x18\x07 \x01(\t\x12\x12\n\nmacaddress\x18\x08 \x01(\t")\n\x15RequestCreateCOHNCert\x12\x10\n\x08override\x18\x01 \x01(\x08"\x16\n\x14RequestClearCOHNCert"\x11\n\x0fRequestCOHNCert"O\n\x10ResponseCOHNCert\x12-\n\x06result\x18\x01 \x01(\x0e2\x1d.open_gopro.EnumResultGeneric\x12\x0c\n\x04cert\x18\x02 \x01(\t",\n\x15RequestSetCOHNSetting\x12\x13\n\x0bcohn_active\x18\x01 \x01(\x08*>\n\x0eEnumCOHNStatus\x12\x16\n\x12COHN_UNPROVISIONED\x10\x00\x12\x14\n\x10COHN_PROVISIONED\x10\x01*\xec\x01\n\x14EnumCOHNNetworkState\x12\x13\n\x0fCOHN_STATE_Init\x10\x00\x12\x14\n\x10COHN_STATE_Error\x10\x01\x12\x13\n\x0fCOHN_STATE_Exit\x10\x02\x12\x13\n\x0fCOHN_STATE_Idle\x10\x05\x12\x1f\n\x1bCOHN_STATE_NetworkConnected\x10\x1b\x12"\n\x1eCOHN_STATE_NetworkDisconnected\x10\x1c\x12"\n\x1eCOHN_STATE_ConnectingToNetwork\x10\x1d\x12\x16\n\x12COHN_STATE_Invalid\x10\x1e'
)
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "cohn_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _ENUMCOHNSTATUS._serialized_start = 537
    _ENUMCOHNSTATUS._serialized_end = 599
    _ENUMCOHNNETWORKSTATE._serialized_start = 602
    _ENUMCOHNNETWORKSTATE._serialized_end = 838
    _REQUESTGETCOHNSTATUS._serialized_start = 50
    _REQUESTGETCOHNSTATUS._serialized_end = 102
    _NOTIFYCOHNSTATUS._serialized_start = 105
    _NOTIFYCOHNSTATUS._serialized_end = 322
    _REQUESTCREATECOHNCERT._serialized_start = 324
    _REQUESTCREATECOHNCERT._serialized_end = 365
    _REQUESTCLEARCOHNCERT._serialized_start = 367
    _REQUESTCLEARCOHNCERT._serialized_end = 389
    _REQUESTCOHNCERT._serialized_start = 391
    _REQUESTCOHNCERT._serialized_end = 408
    _RESPONSECOHNCERT._serialized_start = 410
    _RESPONSECOHNCERT._serialized_end = 489
    _REQUESTSETCOHNSETTING._serialized_start = 491
    _REQUESTSETCOHNSETTING._serialized_end = 535



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
*
Defines the structure of protobuf messages for setting camera control status
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions
DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _EnumCameraControlStatus:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _EnumCameraControlStatusEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_EnumCameraControlStatus.ValueType], builtins.type
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    CAMERA_IDLE: _EnumCameraControlStatus.ValueType
    CAMERA_CONTROL: _EnumCameraControlStatus.ValueType
    "Can only be set by camera, not by app or third party"
    CAMERA_EXTERNAL_CONTROL: _EnumCameraControlStatus.ValueType
    CAMERA_COF_SETUP: _EnumCameraControlStatus.ValueType
    "Set by the camera when it is on the CAH (Camera As a Hub) / COF (Cloud Offload) setup screen"

class EnumCameraControlStatus(_EnumCameraControlStatus, metaclass=_EnumCameraControlStatusEnumTypeWrapper): ...

CAMERA_IDLE: EnumCameraControlStatus.ValueType
CAMERA_CONTROL: EnumCameraControlStatus.ValueType
"Can only be set by camera, not by app or third party"
CAMERA_EXTERNAL_CONTROL: EnumCameraControlStatus.ValueType
CAMERA_COF_SETUP: EnumCameraControlStatus.ValueType
"Set by the camera when it is on the CAH (Camera As a Hub) / COF (Cloud Offload) setup screen"
global___EnumCameraControlStatus = EnumCameraControlStatus

@typing_extensions.final
class RequestSetCameraControlStatus(google.protobuf.message.Message):
    """*
    Set Camera Control Status (as part of Global Behaviors feature)

    This command is used to tell the camera that the app (i.e. External Control) wishes to claim control of the camera.
    This causes the camera to immediately exit most contextual menus and return to the idle screen. Any interaction with
    the camera's physical buttons will cause the camera to reclaim control and update control status accordingly. If the
    user returns the camera UI to the idle screen, the camera updates control status to Idle.

    The entity currently claiming control of the camera is advertised in camera status 114. Information about whether the
    camera is in a contextual menu or not is advertised in camera status 63.

    Response: @ref ResponseGeneric
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    CAMERA_CONTROL_STATUS_FIELD_NUMBER: builtins.int
    camera_control_status: global___EnumCameraControlStatus.ValueType
    "Declare who is taking control of the camera"

    def __init__(self, *, camera_control_status: global___EnumCameraControlStatus.ValueType | None = ...) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["camera_control_status", b"camera_control_status"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["camera_control_status", b"camera_control_status"]
    ) -> None: ...

global___RequestSetCameraControlStatus = RequestSetCameraControlStatus
